/*
$Id: rd_sbs_3d.c,v 1.1 1995/08/09 01:38:51 choudhar Exp $

File Name : rd_sbs_3d.c
Author    : srshah

Entry Point(s) :

(1) passion_read_block_section_3D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_SBS_3D
#define PASSION_RD_SBS_3D

#include <memory.h>
#include <stdio.h>

#include "passion_base.h"
/*
#include "ocad.h"
#include "pfile.h"
*/


/*
  Name : passion_read_block_section_3D

  Function : This function is used to read a block-strided section for 3-D
	array.

  Parameters : 
    Input :
      fd       - LAF descriptor
			Index0    - row index in the array where to put data
			Index1    - column index in the array where to put data
			Index2    - z index in the array where to put data
		  StartIndex0 - start row of the section
		  StartIndex1 - start column of the section
		  StartIndex2 - start z of the section
			EndIndex0 - end row of the section
			EndIndex1 - end column of the section
			EndIndex2 - end z of the section
			stride0  - row stride in the section
			stride1  - column stride in the section
			stride2  - z stride in the section
			block_size0 - block size in the row dimension
			block_size1 - block size in the column dimension
			block_size2 - block size in the z direction
		  elemsize - size of each array element in bytes
			headersize - size of the header
			icla_size - size of the icla in no. of elements
			ocla_size_2 - size of dimension 2 in the OCLA in number of elements
			ocla_size_3 - size of dimension 3 in the OCLA in number of elements
			icla_size_2 - size of dimension 2 in the ICLA in number of elements
			icla_size_3 - size of dimension 3 in the ICLA in number of elements
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

  if no stride in z
	  if no stride in the column
	    if no stride in the row
			  read 3-D unstrided section
		  else
			  you are reading non-consecutive planes of unstrided data
        section start = start index1
        section end   = end   index1
			  array plane     = index0
			  for each plane of interest in the section
			  	compute the offset of the start of the plane from the start of the
				  file and call it new file header
				  new file header = plane * size of ocla plane * size of each element
				                    + file header
				  array index   = (array plane * size of z dim) + index1 
          read the plane as unstrided section of 2-D array 
				  (passion_read_usection_2d)
				  array plane = array plane + 1
			  endfor
		  endif
    else
			consecutive strips of data to be read
			array row = index0
			array col = index1
			for each row of interest
				for each column of interest
			  	compute the offset of the start of the z strip from the start of the
				  file and call it new file header
				  new file header = (row * size of ocla plane + columns * size of
														ocla columns) * size of each element + file header
				  array index   = (array row * size of icla plane + array col *
													 size of icla column) + index2
          read the plane as unstrided section of 1-D array 
				  (passion_read_usection_1d)
				  array col = array col + 1
				endfor
			  array row = array row + 1
			endfor
		endif
  else
		its a general 3-D section
		array row = index0
		array col = index1
		for each row of interest
			for each column of interest
		  	compute the offset of the start of the z strip from the start of the
			  file and call it new file header
			  new file header = (row * size of ocla plane + columns * size of
													ocla columns) * size of each element + file header
			  array index   = (array row * size of icla plane + array col *
												 size of icla column) + index1 * size of icla z dim 
												 + index2
        read the plane as block-strided section of 1-D array 
			  (passion_read_block_section_1d)
			  array col = array col + 1
			endfor
		  array row = array row + 1
		endfor
	endif

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

PREFETCH *passion_pread_block_section_3D(LAF_File fd, char *Array, int Index, int StartIndex, 
				  int u0, int u1, int u2, 
				  int DataLen0, int DataLen1, int DataLen2,
				  int elemsize, int headersize, int icla_size, 
                                  int icla_size_0, int icla_size_1,
				  int ocla_size_0, int ocla_size_1)
{
int i, j, k, l, fetch_size0, fetch_size1;
int NewHeaderSize, ArrayIndex, ArrayPlane, ArrayRow, ArrayCol;
PREFETCH *PREFETCHp, *PREFETCHp1;
int block_size0=1;
int block_size1=1;
int block_size2=1;

	if (DataLen0 <= 0 || DataLen1 <= 0 || DataLen2 <= 0) {
fprintf(stderr,"passion_pread_block_section_3D: DataLen is (less than) zero.\n");
	  return (PREFETCH *)0;
        }


	if ((u0 == 1) || (block_size2 == u0))
	{
	  if ((u1 == ocla_size_0) || (block_size1 == u1))
	  {
	    if ((u2 == ocla_size_0 * ocla_size_1) || (block_size0 == u2))
		  {
			  /* read 3-D unstrided section */
        return passion_pread_usection_3D(fd, Array, Index,  StartIndex, 
				u0, u1, u2, 
				DataLen0, DataLen1, DataLen2,
				elemsize, headersize, 
				icla_size_0, icla_size_1,
				ocla_size_0, ocla_size_1);
		  }
		  else
		  {
			  /* you are reading non-consecutive planes of unstrided data */
			  ArrayPlane     = 0;
			  PREFETCHp      = (PREFETCH *)0;
			  for (i = 0; i < DataLen2; i++)
			  {

				if ((DataLen2 - i + 1) < block_size0) {
					fetch_size0 = DataLen2 - i + 1;
				}
				else {
					fetch_size0 = block_size0;
				}

				for(k = 0; k < fetch_size0; k++) {
			  /* compute the offset of the start of the plane from the start of the
			     file and call it new file header */

				  NewHeaderSize = (i+k) * u2 * elemsize + headersize;
				  ArrayIndex    = ArrayPlane * icla_size_0 * icla_size_1 + Index;
          /* read the plane as unstrided section of 2-D array */
          PREFETCHp1 = passion_pread_usection_2D(fd, Array, ArrayIndex, StartIndex, 
				       u0, u1, DataLen0, DataLen1,
				       elemsize, NewHeaderSize, 
				       icla_size_0, ocla_size_0);
				  if (PREFETCHp1 == (PREFETCH *)0)
				  {
					  #ifdef DEBUG
					  fprintf(stderr, "PASSION : (read prefetch) Read error.\n");
					  #endif
					  PASSION_prefetch_wait(PREFETCHp);
					  return (PREFETCH *)0;
				  }
				  PREFETCHp = passion_mergePREFETCH(PREFETCHp, PREFETCHp1);
				  if (PREFETCHp == (PREFETCH *)0)
				  {
					  #ifdef DEBUG
					  fprintf(stderr, "PASSION : (read prefetch) merge error.\n");
					  #endif
					  return (PREFETCH *)0;
				  }
				  ArrayPlane = ArrayPlane + 1;
					}
			  }
		  }
	  }
    else
	  {
			/* consecutive strips of data to be read */
		  ArrayRow     = 0;
		  PREFETCHp    = (PREFETCH *)0;
		  for (i = 0; i < DataLen2; i++)
		  {
			if ((DataLen2 - i + 1) < block_size0) {
				fetch_size0 = DataLen2 - i + 1;
			}
			else {
				fetch_size0 = block_size0;
			}
			for(k = 0; k < fetch_size0; k++) {
			    ArrayCol     = 0;
			    for (j = 0; j < DataLen1; j++)
			    {
				if ((DataLen1 - j + 1) < block_size1) {
					fetch_size1 = DataLen1 - j + 1;
				}
				else {
					fetch_size1 = block_size1;
				}
				for(l=0; l<fetch_size1; l++) {
			  	/* compute the offset of the start of the z strip from the start of 
						 the file and call it new file header */
			    		NewHeaderSize = ((i+k) * u2 + (j+l) * u1) * elemsize + headersize;
			    		ArrayIndex    = (ArrayRow * icla_size_1 + ArrayCol) * icla_size_0 + Index;
          /* read the plane as unstrided section of 1-D array */
          		PREFETCHp1=passion_pread_usection_1D(fd, Array, ArrayIndex, 
							StartIndex, DataLen0, 
							elemsize, NewHeaderSize);
				  	if (PREFETCHp1 == (PREFETCH *)0)
				  	{
					  fprintf(stderr, "PASSION : (read prefetch) Read error.\n");
					  PASSION_prefetch_wait(PREFETCHp);
					  return (PREFETCH *)0;
				  	}
				  	PREFETCHp = passion_mergePREFETCH(PREFETCHp, PREFETCHp1);
				  	if (PREFETCHp == (PREFETCH *)0)
				  	{
					  fprintf(stderr, "PASSION : (read prefetch) merge error.\n");
					  return (PREFETCH *)0;
				  	}
				  	ArrayCol = ArrayCol + 1;
					} /* end for with l */
				} /* end for with j */
			  ArrayRow = ArrayRow + 1;
				} /* end for with k */
			}
		}
	}
	else
	{
		/* its a general 3-D section */
		/* Too many reads will result */
		#ifdef DEBUG
		fprintf(stderr, "PASSION : (read prefetch) Too many reads.\n");
		#endif
		return (PREFETCH *)0;
	}
	return PREFETCHp;
}

#endif  /* PASSION_RD_SBS_3D */
