
// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize) ;
}

Format BlockRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockRangeRep :: ker_glb(const int glb, int* sub) const {

  *sub = glb % blockSize ;
  return glb / blockSize ;
}

int BlockRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  // `ker_glb' is a global subscript in the kernel range.

  *sub   = 0 ;
  *count = blockSize ;

  return ker_glb * blockSize ;
}

int BlockRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = 1 ;
  return 1 ;
}

int BlockRangeRep :: width() const {return blockSize ; }

StrideRep* BlockRangeRep :: vector(StrideRep* _kernel,
                                   const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int BlockRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb * blockSize + sub ;
}

void BlockRangeRep :: show() {
  AdlibPrintf("BlockRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockRangeRep)

//wjp add 1999.4.8
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GenBlockRangeRep': GenBlock-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* GenBlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new GenBlockSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize, rank, index) ;
}

Format GenBlockRangeRep :: format() const {
  return DIST_GENBLOCK ;
//  return DIST_BLOCK ;//need define a new value --wjp 1999.4.8--
}

//wjp add 1999.5.9 for gen_block
int GenBlockRangeRep :: proc_num()  const {return rank ; }
int* GenBlockRangeRep :: index_()  const {return index ; }

int GenBlockRangeRep :: ker_glb(const int glb, int* sub) const {
    int i,acc_ind,tmp_kerlo;

    acc_ind = 0;
    for (i=0;i<=rank-1;i++) {
      if ((glb >= acc_ind) && (glb < acc_ind + index[i]))
          *sub = glb - acc_ind;
          tmp_kerlo = i; 
      acc_ind += index[i];
    }

    return tmp_kerlo;

//  *sub = glb % blockSize ;
//  return glb / blockSize ;

}

int GenBlockRangeRep :: block(const int ker_glb, int* sub, int* count) const {
    int i,acc_ind;

  // `ker_glb' is a global subscript in the kernel range.

  *sub   = 0 ;
  *count = blockSize ;

    acc_ind = 0;
    for (i=0;i<ker_glb;i++) acc_ind += index[i];
//    AdlibPrintf("GenBlockRangeRep:: block(): sub=%d,count=%d\n",*sub,*count);

    return acc_ind ;
//  return ker_glb * blockSize ;
}

int GenBlockRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = 1 ;
  return 1 ;
}

int GenBlockRangeRep :: width() const {return blockSize ; }

StrideRep* GenBlockRangeRep :: vector(StrideRep* _kernel,
                                   const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int GenBlockRangeRep :: glb(const int sub, const int ker_glb) const {
    int i,acc_ind;

    acc_ind = 0;
    for (i=0;i<ker_glb;i++) acc_ind += index[i];

    return acc_ind + sub ;
//  return ker_glb * blockSize + sub ;
}

void GenBlockRangeRep :: show() {
  AdlibPrintf("GenBlockRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(GenBlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, blockSize) ;
}

RangeRep* BlockSubRangeRep :: tem() {return parent ; }
int BlockSubRangeRep :: str() const {return stride ; }
int BlockSubRangeRep :: bas() const {return base ; }

Format BlockSubRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockSubRangeRep :: ker_glb(const int glb, int* sub) const {
  const int tem = base + stride * glb ;

//AdlibPrintf("BlockSubRangeRep:: ker_glb(): base=%d,stride=%d,glb=%d,tem=%d,blockSize=%d\n",base,stride,glb,tem,blockSize);
  *sub = tem % blockSize ;
  return tem / blockSize ;
}

//int BlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {
//
//  const int blockStartOffset  = ker_glb * blockSize - base ;
//  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;
//
//  int i_lo, i_hi ;
//
//  if(stride > 0) {
//    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
//    i_hi = min(blockFinishOffset / stride, extent - 1) ;
//  }
//  else {
//    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
//    i_hi = min(blockStartOffset / stride, extent - 1) ;
//  }
//
//  *sub   = base + stride * i_lo ;
//  *count = max(i_hi - i_lo + 1, 0) ;
//
//  return i_lo ;
//}

int BlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

//AdlibPrintf("BlockSubRangeRep:: block(): ker_glb=%d,ker_lo=%d,ker_hi=%d\n",ker_glb,ker_lo,ker_hi);
  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    const int blockStart = ker_glb * blockSize ;

    const int blockStartOffset  = blockStart - base ;
    const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

    int i_hi, i_lo ;
    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }
   
    *sub   = base + stride * i_lo - blockStart ;
    *count = i_hi - i_lo + 1 ;

//AdlibPrintf("BlockSubRangeRep:: block(): sub=%d,count=%d\n",*sub,*count);
    return i_lo ;
  }
  else {
    *sub   = base ;
    *count = 0 ;

    return 0 ;
  }
}

int BlockSubRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = stride ;
  return 1 ;
}

RangeRep* BlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {

  // Do we need the reversal of the subkernel for -ve stride?

  if(stride > 0) {
    *ker_bas = ker_lo ;
    *ker_str = 1 ;
  }
  else {
    *ker_bas = ker_hi ;
    *ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, *ker_bas, *ker_str) ;
}

//RangeRep* BlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {
//
//  // Tempting to reverse stride of subkernel if `stride' is negative.
//
//  *ker_bas = ker_lo ;
//  *ker_str = 1 ;
//
//  return kernel->subrng(ker_hi - ker_lo + 1, ker_lo, 1) ;
//}

int BlockSubRangeRep :: width() const {return blockSize ; }

StrideRep* BlockSubRangeRep :: vector(StrideRep* _kernel,
                                      const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int BlockSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (ker_glb * blockSize + sub - base) / stride ;
}

void BlockSubRangeRep :: show() {
  AdlibPrintf("BlockSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockSubRangeRep)

//wjp add 1999.4.8
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GenBlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* GenBlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new GenBlockSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, blockSize, rank, index) ;
}

RangeRep* GenBlockSubRangeRep :: tem() {return parent ; }
int GenBlockSubRangeRep :: str() const {return stride ; }
int GenBlockSubRangeRep :: bas() const {return base ; }

Format GenBlockSubRangeRep :: format() const {
  return DIST_GENBLOCK ;
//  return DIST_BLOCK ;//need define a new value --wjp 1999.4.12--
}

//wjp add 1999.5.9 for gen_block
int GenBlockSubRangeRep :: proc_num()  const {return rank ; }
int* GenBlockSubRangeRep :: index_()  const {return index ; }

int GenBlockSubRangeRep :: ker_glb(const int glb, int* sub) const {
    const int tem = base + stride * glb ;
    int i,acc_ind,tmp_kerlo;
//AdlibPrintf("GenBlockSubRangeRep:: ker_glb(): base=%d,stride=%d,glb=%d,tem=%d,blockSize=%d\n",base,stride,glb,tem,blockSize);

    acc_ind = 0;
    for (i=0;i<=rank-1;i++) {
      if ((tem >= acc_ind) && (tem < acc_ind + index[i])) {
          *sub = tem - acc_ind;
          tmp_kerlo = i; 
      }
      acc_ind += index[i];
    }

    if (tem >=acc_ind) {
          *sub = tem - acc_ind;
          tmp_kerlo = rank; 
    }

    return tmp_kerlo;


//  *sub = tem % blockSize ;
//  return tem / blockSize ;
}

//int BlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {
//
//  const int blockStartOffset  = ker_glb * blockSize - base ;
//  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;
//
//  int i_lo, i_hi ;
//
//  if(stride > 0) {
//    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
//    i_hi = min(blockFinishOffset / stride, extent - 1) ;
//  }
//  else {
//    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
//    i_hi = min(blockStartOffset / stride, extent - 1) ;
//  }
//
//  *sub   = base + stride * i_lo ;
//  *count = max(i_hi - i_lo + 1, 0) ;
//
//  return i_lo ;
//}

int GenBlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

//  AdlibPrintf("GenBlockSubRangeRep:: block(): ker_glb=%d,ker_lo=%d,ker_hi=%d\n",ker_glb,ker_lo,ker_hi);
  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    int i,tmp_s;
    tmp_s = 0; 
    for(i=0;i<ker_glb;i++) tmp_s += index[i];
    const int blockStart = tmp_s ;
//  AdlibPrintf("GenBlockSubRangeRep:: block(): tmp_s=%d\n",tmp_s);
//    const int blockStart = ker_glb * blockSize ;

    const int blockStartOffset  = blockStart - base ;
//   const int blockFinishOffset = blockStartOffset + blockSize - 1 ;
    const int blockFinishOffset = blockStartOffset + index[ker_glb] - 1 ;

    int i_hi, i_lo ;
    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }
   
    *sub   = base + stride * i_lo - blockStart ;
    *count = i_hi - i_lo + 1 ;

//   AdlibPrintf("GenBlockSubRangeRep:: block(): sub=%d,count=%d\n",*sub,*count);

    return i_lo ;
  }
  else {
    *sub   = base ;
    *count = 0 ;

    return 0 ;
  }
}

int GenBlockSubRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = stride ;
//  AdlibPrintf("GenBlockSubRangeRep:: block_step(): sub_stp=%d\n",*sub_stp);

  return 1 ;
}

RangeRep* GenBlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {

  // Do we need the reversal of the subkernel for -ve stride?

  if(stride > 0) {
    *ker_bas = ker_lo ;
    *ker_str = 1 ;
  }
  else {
    *ker_bas = ker_hi ;
    *ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, *ker_bas, *ker_str) ;
}

//RangeRep* BlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {
//
//  // Tempting to reverse stride of subkernel if `stride' is negative.
//
//  *ker_bas = ker_lo ;
//  *ker_str = 1 ;
//
//  return kernel->subrng(ker_hi - ker_lo + 1, ker_lo, 1) ;
//}

int GenBlockSubRangeRep :: width() const {return blockSize ; }

StrideRep* GenBlockSubRangeRep :: vector(StrideRep* _kernel,
                                      const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int GenBlockSubRangeRep :: glb(const int sub, const int ker_glb) const {
    int i,acc_ind;

    acc_ind = 0;
    for (i=0;i<ker_glb;i++) acc_ind += index[i];

    return (acc_ind + sub - base) / stride ;
//  return (ker_glb * blockSize + sub - base) / stride ;
}

void GenBlockSubRangeRep :: show() {
  AdlibPrintf("GenBlockSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(GenBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockRangeRep', `PackedBlockSubRangeRep':            //
//      Block-distributed ranges yielding arrays with ``packed''       //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent, _base, _stride,
                                    this, kernel, blockSize) ;
}

RangeRep* PackedBlockSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent,
				    base + stride * _base, stride * _stride,
				    parent, kernel, blockSize) ;
}

StrideRep* PackedBlockSubRangeRep :: vector(StrideRep* _kernel,
                                            const int _length) const {
  return new PackedStrideRep(_kernel, _length, packing) ;
}

int PackedBlockSubRangeRep :: width() const {
  return (blockSize + packing - 1) / packing ;
}

MEMORY_POOL_DEFN(PackedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedStrideRep'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int PackedStrideRep :: disp(const int sub) const {
  return length * (sub / packing) ;
}

int PackedStrideRep :: disp_step(const int sub_stp) const {
  return length * (sub_stp / packing) ;
}

MEMORY_POOL_DEFN(PackedStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedBlockRangeRep :: subrng(const int _extent,
                                          const int _base, const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent, _base, _stride, this,
                                      kernel, blockSize, wlo, whi) ;
}

int ExtendedBlockRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedBlockRangeRep :: vector(StrideRep* _kernel,
                                           const int _length) const {
  return new ExtendedStrideRep(_kernel, _length, wlo) ;
}


RangeRep* ExtendedBlockSubRangeRep :: subrng(const int _extent,
                                             const int _base,
                                             const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent,
				      base + stride * _base, stride * _stride,
				      parent, kernel, blockSize, wlo, whi) ;
}

int ExtendedBlockSubRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedBlockSubRangeRep :: vector(StrideRep* _kernel,
                                              const int _length) const {
  return new ExtendedStrideRep(_kernel, _length, wlo) ;
}


MEMORY_POOL_DEFN(ExtendedBlockRangeRep)

MEMORY_POOL_DEFN(ExtendedBlockSubRangeRep)


//wjp add 1999.5.9 for gen_block
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedGenBlockRangeRep', `ExtendedgenBlockSubRangeRep':  //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedGenBlockRangeRep :: subrng(const int _extent,
                                          const int _base, const int _stride) {
  return new ExtendedGenBlockSubRangeRep(_extent, _base, _stride, this,
                                      kernel, blockSize, wlo, whi, rank, index) ;
}

int ExtendedGenBlockRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedGenBlockRangeRep :: vector(StrideRep* _kernel,
                                           const int _length) const {
  return new ExtendedGenStrideRep(_kernel, _length, wlo) ;
}


RangeRep* ExtendedGenBlockSubRangeRep :: subrng(const int _extent,
                                             const int _base,
                                             const int _stride) {
  return new ExtendedGenBlockSubRangeRep(_extent,
				      base + stride * _base, stride * _stride,
				      parent, kernel, blockSize, wlo, whi, rank, index) ;
}

int ExtendedGenBlockSubRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedGenBlockSubRangeRep :: vector(StrideRep* _kernel,
                                              const int _length) const {
  return new ExtendedGenStrideRep(_kernel, _length, wlo) ;
}


MEMORY_POOL_DEFN(ExtendedGenBlockRangeRep)

MEMORY_POOL_DEFN(ExtendedGenBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedStrideRep'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int ExtendedStrideRep :: disp(const int sub) const {
  return length * (sub + wlo) ;
}

MEMORY_POOL_DEFN(ExtendedStrideRep)


//wjp add 1999.5.9 for gen_block
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedGenStrideRep'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int ExtendedGenStrideRep :: disp(const int sub) const {
  return length * (sub + wlo) ;
}

MEMORY_POOL_DEFN(ExtendedGenStrideRep)


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)
//
//   Check if reversal in `BlockSubRangeRep :: subker()' can be omitted.

