
// Schedules for deterministic communication

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "Procs.h"
#include "threads.h"
#include "adtags.h"
#include "schedule.h"
#include "MessSchedule.h"

int InverseRemp = 0;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `MessSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

MessSchedule :: MessSchedule(const int _elementLen) : elementLen(_elementLen),
		locStatus(IN_CORE), srvStatus(IN_CORE)
{
  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

  locIdxs = new IndexList[nProcs] ;
  srvIdxs = new IndexList[nProcs] ;
/*
  locIdxs = (IndexList *)malloc(sizeof(IndexList) * nProcs);
  srvIdxs = (IndexList *)malloc(sizeof(IndexList) * nProcs);
*/
}

MessSchedule :: MessSchedule(const int _elementLen, Status _locStatus, Status _srvStatus) : elementLen(_elementLen) {
  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

  locIdxs = new IndexList [nProcs] ;
  srvIdxs = new IndexList [nProcs] ;
/*
  locIdxs = (IndexList *)malloc(sizeof(IndexList) * nProcs);
  srvIdxs = (IndexList *)malloc(sizeof(IndexList) * nProcs);
*/

    // add by Zheng Gengbin
  locStatus = _locStatus;
  srvStatus = _srvStatus;
}

void MessSchedule :: build() {

  // Set `locList' and `srvList' to hop-ordered lists of non-empty messages,
  // and allocate contiguous buffers big enough to hold largest messages.

  int locBufSize = 0, srvBufSize = 0 ;
  for(int hop = 0 ; hop < nProcs ; hop++) {
    {
      const int remId = (locId + hop) % nProcs ;
      int size = locIdxs [remId].size() ;
      if(size) {
	locList.add(remId) ;
        if(size > locBufSize) locBufSize = size ;
      }
    }
    {
      const int remId = (locId - hop + nProcs) % nProcs ;
      int size = srvIdxs [remId].size() ;
      if(size) {
	srvList.add(remId) ;
        if(size > srvBufSize) srvBufSize = size ;
      }
    }
  }
  locBuf = new char [locBufSize * elementLen] ;  // *** potential word-
  srvBuf = new char [srvBufSize * elementLen] ;  //     alignment problem ***
}

MessSchedule :: ~MessSchedule() {
  delete [] srvBuf ;
  delete [] locBuf ;

  delete [] srvIdxs ;
  delete [] locIdxs ;
}

void MessSchedule :: gather(void* locBase, void* srvBase) {

  // Spawn server to send messages in `srvList'...

  SendServer sndServer(srvList.size(), srvList.data(),
                       srvBuf, srvIdxs, srvBase, elementLen, srvStatus) ;  // Zheng Gengbin

  // Receive messages in `locList'...

  const int numMes = locList.size() ;
  int* remIds      = locList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = locIdxs + remId ;
    const int count = indexVec->size() ;
    int*    offsets = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(locBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

      // modified by Zheng Gengbin
    if (locStatus == IN_CORE)
    	scatterOffBuf((char*) locBase, offsets, count, locBuf, elementLen) ;
    else {
	PFILE *locPfile = (PFILE*) locBase;
    	oc_scatterOffBuf(locPfile , offsets, count, locBuf, elementLen) ;
    }
  }

  // Wait for `sndServer' to terminate.

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: scatter(void* locBase, void* srvBase) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

      // modified by Zheng Gengbin
    if (srvStatus == IN_CORE)
        scatterOffBuf((char*) srvBase, offsets, count, srvBuf, elementLen) ;
    else {
	PFILE *srvPfile = (PFILE*) srvBase;
    	oc_scatterOffBuf(srvPfile , offsets, count, srvBuf, elementLen) ;
    }
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: combine(void* locBase, void* srvBase,
                             void (*op)(void*, void*)) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    char* buf = srvBuf ;
    for(int i = 0 ; i < count ; i++) {
      (*op)((char*) srvBase + elementLen * offsets [i], buf) ;
      buf += elementLen ;
    }
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `SendServer' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

SendServer :: SendServer(const int _numMes, int* _remIds,
			 char* _sendBuf, IndexList* _sendIdxs,
			 void* _sendBase, int _elementLen, Status _status) :
                     numMes(_numMes), remIds(_remIds),
                     sendBuf(_sendBuf), sendIdxs(_sendIdxs),
                     sendBase(_sendBase), elementLen(_elementLen),
		     status(_status) {
  mes = 0 ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

SendServer :: ~SendServer() {}

void SendServer :: transition(MPI_Status*) {
  mes++ ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

int SendServer :: active() const {
  return mes < numMes ;
}

void SendServer :: initSend(const int remId) {
  IndexList* indexVec = sendIdxs + remId ;
  const int count = indexVec->size() ;
  int*    offsets = indexVec->data() ;

    // Zheng Gengbin
  if (status == IN_CORE)
  	gatherOffBuf((char*) sendBase, offsets, count, sendBuf, elementLen) ;
  else {
	PFILE *sendPfile = (PFILE*) sendBase;
  	oc_gatherOffBuf(sendPfile, offsets, count, sendBuf, elementLen) ;
  }

  MPI_Isend(sendBuf, count * elementLen, MPI_CHAR,
	    remId, DAT, MPI_COMM_WORLD, event) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockMessSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

/*
old RUN code 
void BlockMessSchedule :: run(List<MessQueue>* dstMessages,
                              List<MessQueue>* srcMessages,
                              MultiBlockList* dstCopies,
                              MultiBlockList* srcCopies,
                              void* dstBase, void* srcBase) {

  // Spawn server to send messages in `srcMessages'...

  BlockSendServer sndServer(srcMessages, srcBase, elementLen) ;

  // Do local copying...

  MultiBlock dstBlock = dstCopies->data() ;
  MultiBlock srcBlock = srcCopies->data() ;

  for(int i = 0 ; i < cpyNum ; i++) {
    copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen) ;
    nextBlock(dstBlock) ;
    nextBlock(srcBlock) ;
  }

  // Receive messages in `dstList'...

  List<MessQueue>* record = dstMessages ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    MPI_Status status ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;
      const int size = message->size ;

      if(num == 1 && block [MBLK_CTG]) {
        char* blockBase  = (char*) dstBase + elementLen * block [MBLK_OFF] ;

	MPI_Irecv(blockBase, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        nextBlock(block) ;
      }
      else {
	MPI_Irecv(statBuf, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        char* ptr = (char*) statBuf ;

        for(int j = 0 ; j < num ; j++) {
          scatterMblkBuf(dstBase, block, ptr, elementLen) ;

          ptr += block [MBLK_SIZ] ;
          nextBlock(block) ;
        }
      }

      message++ ;
    }
    
    record = record->next ;
  }

  // Wait for `sndServer' to terminate.

  while(sndServer.active())
    Server :: waitServerTransition() ;
}
*/

#ifdef HAVE_PTHREAD
// for pthread
void* localCopy(void **arg)
{
  SparseArray<MessQueue>* dstMemCopy = (SparseArray<MessQueue> *)arg[0];
  SparseArray<MessQueue>* srcMemCopy = (SparseArray<MessQueue> *)arg[1];
  void* dstBase = arg[2];
  void* srcBase = arg[3]; 
  int elementLen = *(int *)arg[4];
  {
  List<MessQueue>* srcRecord = srcMemCopy->list ;
  List<MessQueue>* dstRecord = dstMemCopy->list ;
  while(dstRecord) {
    const int numMes = dstRecord->messages.size() ;

    Mess*    message = dstRecord->messages.data() ;

    MultiBlock dstBlock = dstRecord->blocks.data() ;
    MultiBlock srcBlock = srcRecord->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;
      const int size = message->size ;

      for (int i = 0; i < num; i++) {
      	oc_copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen);
	nextBlock(dstBlock); 
	nextBlock(srcBlock); 
      }

      message++ ;
    }
    
    dstRecord = dstRecord->next ;
    srcRecord = srcRecord->next ;
  }
  }   // end of block copy
  return(NULL);
}
#endif

// add by Zheng Gengbin
// new RUN for both IN_CORE & OUT_OF_CORE
void BlockMessSchedule :: run(List<MessQueue>* dstMessages,
                              List<MessQueue>* srcMessages,
                              SparseArray<MessQueue>* dstMemCopy,
                              SparseArray<MessQueue>* srcMemCopy,
                              void* dstBase, void* srcBase) {

  // Spawn server to send messages in `srcMessages'...

  BlockSendServer sndServer(srcMessages, srcBase, elementLen) ;

#ifdef HAVE_PTHREAD
  {
  pthread_t thread;
  char *status;
  void *args[]={ dstMemCopy, srcMemCopy, dstBase, srcBase, &elementLen };
  if (pthread_create(&thread, NULL, localCopy, (void*)&args) )
  {
	AdlibPrintf("Error: creating new thread\n");
  }
  pthread_yield();
  pthread_join(thread, (void **)(&status));
  }
#else
  // Do local copying...
  {
  List<MessQueue>* srcRecord = srcMemCopy->list ;
  List<MessQueue>* dstRecord = dstMemCopy->list ;
  while(dstRecord) {
    const int numMes = dstRecord->messages.size() ;

    Mess*    message = dstRecord->messages.data() ;

    MultiBlock dstBlock = dstRecord->blocks.data() ;
    MultiBlock srcBlock = srcRecord->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;
      const int size = message->size ;

      for (int i = 0; i < num; i++) {
      	oc_copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen);
	nextBlock(dstBlock); 
	nextBlock(srcBlock); 
      }

      message++ ;
    }
    
    dstRecord = dstRecord->next ;
    srcRecord = srcRecord->next ;
  }
  }   // end of block copy
#endif

  // Receive messages in `dstList'...

  List<MessQueue>* record = dstMessages ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    MPI_Status status ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;
      const int size = message->size ;

      if(num == 1 && block [MBLK_CTG] && block[MBLK_STA] == IN_CORE ) {
        char* blockBase  = (char*) dstBase + elementLen * block [MBLK_OFF] ;

	MPI_Irecv(blockBase, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        nextBlock(block) ;
      }
      else {

        ProfBeginRecv()

	MPI_Irecv(statBuf, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        ProfEndRecv()

        char *ptr = (char*) statBuf ;

        for(int j = 0 ; j < num ; j++) {
          if (block[MBLK_STA] == IN_CORE)
             scatterMblkBuf(dstBase, block, ptr, elementLen) ;
          else {
	     PFILE *dstPfile = (PFILE*) dstBase;
             oc_scatterMblkBuf(dstPfile, block, ptr, elementLen) ;
	  }

          ptr += block [MBLK_SIZ] ;
          nextBlock(block) ;
        }

      }

      message++ ;
    }
    
    record = record->next ;
  }

  // Wait for `sndServer' to terminate.
  ProfBeginSend()

  while(sndServer.active())
    Server :: waitServerTransition() ;

  ProfEndSend()
}

void BlockMessSchedule :: show() {

  char buf [1024] ;

  // Local copying...

  MultiBlock dstBlock = locCopies.data() ;
  MultiBlock srcBlock = srvCopies.data() ;

  for(int i = 0 ; i < cpyNum ; i++) {
    const int rank      = dstBlock [MBLK_RNK] ;
    int* dstStrs        = dstBlock + MBLK_STR ;
    int* srcStrs        = srcBlock + MBLK_STR ;
    int* exts           = dstStrs  + rank ;

    char* ptr = buf ;

    ptr += sprintf(ptr, "Copy (") ;
    if(rank) {
      ptr += sprintf(ptr, "%d", exts [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", exts [j]) ;
    }
    ptr += sprintf(ptr, ") * %d from %d + (", elementLen, srcBlock [MBLK_OFF]) ;
    if(rank) {
      ptr += sprintf(ptr, "%d", srcStrs [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", srcStrs [j]) ;
    }
    ptr += sprintf(ptr, ") to %d + (", dstBlock [MBLK_OFF]) ;
    if(rank) {
      ptr += sprintf(ptr, "%d", dstStrs [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", dstStrs [j]) ;
    }
    ptr += sprintf(ptr, ")") ;
   
    AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

    nextBlock(dstBlock) ;
    nextBlock(srcBlock) ;
  }

  List<MessQueue>* record ;

  // Messages to be sent...

  record = srvList ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;

      for(int j = 0 ; j < num ; j++) {
	const int rank   = block [MBLK_RNK] ;
	int* strs        = block + MBLK_STR ;
	int* exts        = strs  + rank ;

	char* ptr = buf ;

	ptr += sprintf(ptr, "Send (") ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", exts [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", exts [k]) ;
	}
	ptr += sprintf(ptr, ") * %d from %d + (",
               elementLen, block [MBLK_OFF]) ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", strs [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", strs [k]) ;
	}
	ptr += sprintf(ptr, ") to proc %d", remId) ;
       
	AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

	nextBlock(block) ;
      }

      message++ ;
    }
    
    record = record->next ;
  }

  // Messages to be recieved...

  record = locList ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;

      for(int j = 0 ; j < num ; j++) {
	const int rank   = block [MBLK_RNK] ;
	int* strs        = block + MBLK_STR ;
	int* exts        = strs  + rank ;

	char* ptr = buf ;

	ptr += sprintf(ptr, "Recv (") ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", exts [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", exts [k]) ;
	}
	ptr += sprintf(ptr, ") * %d to %d + (", elementLen, block [MBLK_OFF]) ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", strs [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", strs [k]) ;
	}
	ptr += sprintf(ptr, ") from proc %d", remId) ;
       
	AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

	nextBlock(block) ;
      }

      message++ ;
    }
    
    record = record->next ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockSendServer' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

//wjp add 1998-3-8
BlockSendServer :: BlockSendServer(List<MessQueue>* sendList,
                                   void* _sendBase, int _elementLen) :
		         record(sendList),
			 sendBase(_sendBase),
                         elementLen(_elementLen) {
  if(record) {
    remId   = record->remId ;
    numMes  = record->messages.size() ;

    mes     = 0 ;
    message = record->messages.data() ;

    block   = record->blocks.data() ;

    initSend() ;
  }
}

void BlockSendServer :: transition(MPI_Status*) {
  mes++ ;
  if(mes < numMes) {
    message++ ;

    initSend() ;
  }
  else {
    record = record->next ;

    if(record) {
      remId   = record->remId ;
      numMes  = record->messages.size() ;

      mes     = 0 ;
      message = record->messages.data() ;

      block   = record->blocks.data() ;

      initSend() ;
    }
  }
}

void* BlockSendServer :: active() const {
  return record ;
}

void BlockSendServer :: initSend() {
  const int num  = message->num  ;
  const int size = message->size ;

     // only IN_CORE case will do this branch
  if(num == 1 && block [MBLK_CTG] && block[MBLK_STA] == IN_CORE) {
    char* blockBase  = (char*) sendBase + elementLen * block [MBLK_OFF] ;
    nextBlock(block) ;

    ProfBeginSend()
    MPI_Isend(blockBase, size, MPI_CHAR,
	      remId, DAT, MPI_COMM_WORLD, event) ;
    ProfEndSend()
  }
  else {
    char* ptr = (char*) sendBuf ;

    for(int j = 0 ; j < num ; j++) {
      if (block[MBLK_STA] == IN_CORE)
         gatherMblkBuf(sendBase, block, ptr, elementLen) ;
      else {
         PFILE *sendPfile = (PFILE *)sendBase;
         oc_gatherMblkBuf(sendPfile, block, ptr, elementLen) ;
      }

      ptr += block [MBLK_SIZ] ;
      nextBlock(block) ;
    }

    ProfBeginSend()

    MPI_Isend(sendBuf, size, MPI_CHAR, remId, DAT, MPI_COMM_WORLD, event) ;

    ProfEndSend()
  }
}


// ----------------------------------------------------------------------
// Things to do
//
//     Change `MessSchedule' to use statically declared buffer.
//     General reorganizations to look more like modern `BlockMessSchedule'.
//
//     Replace `scatterOffBuf' by a virtual function, and unify implementations
//     of `gather', `scatter', `combine'.
//
//     Could `run' be a global function (as it was in the bad old
//     days), or a member of some class other than `BlockDataSchedule'?

