
// Common utilities for constructing schedules.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "passion_base.h"  // add by Zheng Gengbin
#include "threads.h"
#include "block.h"
#include "adtags.h"
#include "schedule.h"

int storageMode;
int defaultStorageMode;
int reduceMode;
int runtimeDebug;

align_t statBuf [(STATIC_BUFFER_SIZE + ALIGN_SZ - 1) / ALIGN_SZ] ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Add block to message queue.                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

static void splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset, int* strs, int contig,
                              int* exts, int size, int rank, Status status, 
			      int iclaSize, int storage) ;

inline void newMessage(MessList &messages, MultiBlockList &blocks,
		       const int offset, const int* strs, const int contig,
		       const int* exts, const int size, const int rank, 
		       int storage, Status status = IN_CORE) {
  Mess* message = messages.add() ;

  message->size  = size ;
  message->num   = 1 ;

  addBlock(blocks, offset, strs, contig, size, exts, rank, storage, status) ;
}

void addMessageBlock(MessList &messages, MultiBlockList &blocks,
		     const int offset, int* strs, const int contig,
		     const int size, int* exts, const int rank, 
		     Status status , const int iclaSize, int storage) {

//AdlibPrintf("addMessageBlock:rnk:%d,offset:%d,ext:%d,%d,store:%d, str:%d,%d \n",rank,offset,exts[0],exts[1],storage,strs[0],strs[1]);
  if (storage == -1) storage = defaultStorageMode;
//AdlibPrintf("BLOCK %d \n");
     // add by Zheng Gengbin    // the message size must smaller than iclaSize
  if(size <= iclaSize) {
    if(messages.size() == 0)
      newMessage(messages, blocks, offset, strs, contig, exts, size, rank, storage, status) ;
    else {
      Mess* message = messages.last() ;

      int newSize = message->size + size ;
      if(newSize <= iclaSize) {
	message->size = newSize ;
	message->num++ ;
	addBlock(blocks, offset, strs, contig, size, exts, rank, storage, status) ;
      }
      else
	newMessage(messages, blocks, offset, strs, contig, exts, size, rank, storage, status) ;
    }
  }
  else
    splitMessageBlock(messages, blocks, offset, strs, contig,
                      exts, size, rank, status, iclaSize, storage) ;
}

/*
void addMessageBlock(MessList &messages, MultiBlockList &blocks,
		     const int offset, int* strs, const int contig,
		     const int size, int* exts, const int rank, Status status , 
		     const int iclaSize) 
{

  int storage = defaultStorageMode;

     // add by Zheng Gengbin    // the message size must smaller than iclaSize
  if(size <= iclaSize) {
    if(messages.size() == 0)
      newMessage(messages, blocks, offset, strs, contig, exts, size, rank, storage, status) ;
    else {
      Mess* message = messages.last() ;

      int newSize = message->size + size ;
      if(newSize <= iclaSize) {
	message->size = newSize ;
	message->num++ ;
	addBlock(blocks, offset, strs, contig, size, exts, rank, storage, status) ;
      }
      else
	newMessage(messages, blocks, offset, strs, contig, exts, size, rank, storage, status) ;
    }
  }
  else
    splitMessageBlock(messages, blocks, offset, strs, contig,
                      exts, size, rank, status, iclaSize, storage) ;
}
*/

static void Col_splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset, int* strs, int contig,
                              int* exts, int size, int rank, Status status, int iclaSize) {

  // Called iff size > STATIC_BUFFER_SIZE.

  // Block is split into a series of messages, with one sub-block
  // per message.

  if(rank == 0)
    newMessage(messages, blocks, offset, strs, contig, exts, size, rank, 
               COLUMN_MAJOR, status) ;
        // Single element larger than static buffer size.
        // Don't try to split individual elements---if a buffer is needed,
        // will have to allocate it dynamically at schedule execution time.

  else {
    int subRank = rank - 1 ;
    int count   = exts [subRank] ;
    int stride  = strs [subRank] ;
    int subSize = size / count ;
    if(subSize <= iclaSize) {

      // Slice leading dimension

      int sliceWidth = iclaSize / subSize ;
      int sliceSize  = sliceWidth * subSize ;
      int sliceStep  = sliceWidth * stride ;
      int numWholeSlices = count / sliceWidth ;

      exts [subRank] = sliceWidth ;

      for(int i = 0 ; i < numWholeSlices ; i++) {
	newMessage(messages, blocks, offset, strs, contig,
                   exts, sliceSize, rank, COLUMN_MAJOR, status) ;
	offset += sliceStep ;
      }

      int residue = count - numWholeSlices * sliceWidth ;
      if(residue) {
	exts [subRank] = residue ;
	newMessage(messages, blocks, offset, strs, contig,
		   exts, residue * subSize, rank, COLUMN_MAJOR, status) ;
      }

      exts [subRank] = count ;  // restore to original value.
    }
    else {

      // Reduced rank sub-blocks still too large.  Recurse.

      for(int i = 0  ; i < count ; i++) {
	Col_splitMessageBlock(messages, blocks, offset, strs, contig,
                          exts, subSize, subRank, status, iclaSize) ;
	offset += stride ;
      }
    }
  }
}

static void Row_splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset, int* strs, int contig,
                              int* exts, int size, int rank, int d, Status status, int iclaSize) {

  // Called iff size > STATIC_BUFFER_SIZE.

  // Block is split into a series of messages, with one sub-block
  // per message.

  if(rank == 0)
    newMessage(messages, blocks, offset, strs, contig, exts, size, rank, 
               ROW_MAJOR, status) ;
        // Single element larger than static buffer size.
        // Don't try to split individual elements---if a buffer is needed,
        // will have to allocate it dynamically at schedule execution time.

  else {
    int subRank = d -1;
    int count   = exts [subRank] ;
    int stride  = strs [subRank] ;
    int subSize = size / count ;
    if(subSize <= iclaSize) {

      // Slice leading dimension

      int sliceWidth = iclaSize / subSize ;
      int sliceSize  = sliceWidth * subSize ;
      int sliceStep  = sliceWidth * stride ;
      int numWholeSlices = count / sliceWidth ;

      exts [subRank] = sliceWidth ;

      for(int i = 0 ; i < numWholeSlices ; i++) {
	newMessage(messages, blocks, offset, strs + subRank, contig,
                   exts + subRank, sliceSize, rank - subRank, ROW_MAJOR, status);
	offset += sliceStep ;
      }

      int residue = count - numWholeSlices * sliceWidth ;
      if(residue) {
	exts [subRank] = residue ;
	newMessage(messages, blocks, offset, strs + subRank, contig,
		   exts + subRank, residue * subSize, rank - subRank, 
                   ROW_MAJOR, status) ;
      }

      exts [subRank] = count ;  // restore to original value.
    }
    else {

      // Reduced rank sub-blocks still too large.  Recurse.

      for(int i = 0  ; i < count ; i++) {
	Row_splitMessageBlock(messages, blocks, offset, strs, contig,
                          exts, subSize, rank, d+1, status, iclaSize) ;
	offset += stride ;
      }
    }
  }
}

static void splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset, int* strs, int contig,
                              int* exts, int size, int rank, Status status, 
			      int iclaSize, int storage) 
{
  if (defaultStorageMode == COLUMN_MAJOR || reduceMode == 1)
	Col_splitMessageBlock(messages, blocks, offset, strs, contig,
                              exts, size, rank, status, iclaSize);
  else
	Row_splitMessageBlock(messages, blocks, offset, strs, contig,
                              exts, size, rank, 1, status, iclaSize);
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Add a pair of blocks to message queue.                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// (Hackishly) modified version of the standard code above.

// This code was added specially to support `BlockRedx2Schedule', which is
// used to implement the `dim' form of `maxloc', `minloc'.

// Note that the `MBLK_SIZ' element of each block is rounded up by
// `ALIGN_UP' to avoid word alignment problems when two different types
// may be mixed in the same message.

//xh
static void splitMessageBlocks(MessList &messages, MultiBlockList &blocks,
		               int offset1, int* strs1, int contig1, int size1,
		               int offset2, int* strs2, int contig2, int size2,
                               int* exts, int rank,Status status,int iclaSize,
                               int storage) ;
//xh
inline void newMessage(MessList &messages, MultiBlockList &blocks,
      const int offset1, const int* strs1, const int contig1, const int size1,
      const int offset2, const int* strs2, const int contig2, const int size2,
      const int* exts, const int rank, const int size,
      int storage,Status status=IN_CORE){ 
  Mess* message = messages.add() ;

  message->size  = size ;
  message->num   = 1 ;

  addBlock(blocks, offset1, strs1, contig1, size1, exts, rank, storage,status ) ;
  addBlock(blocks, offset2, strs2, contig2, size2, exts, rank, storage,status ) ;
}

void addMessageBlocks(MessList &messages, MultiBlockList &blocks,
      const int offset1, int* strs1, const int contig1, const int size1,
      const int offset2, int* strs2, const int contig2, const int size2,
      int* exts, const int rank,Status status,const int iclaSize,int storage) {

//xh
  if (storage == -1) storage = defaultStorageMode;

//  const int a_size1 = ALIGN_UP(size1) ;
//  const int a_size2 = ALIGN_UP(size2) ;
//  const int size = a_size1 + a_size2 ;

   const int size = size1 + size2 ;

  if(size <= iclaSize) {
    if(messages.size() == 0)
      newMessage(messages, blocks,
                 offset1, strs1, contig1, size1,
                 offset2, strs2, contig2, size2, 
                 exts, rank, size,storage, status) ;
    else {
      Mess* message = messages.last() ;

      int newSize = message->size + size ;
      if(newSize <= iclaSize) {
	message->size = newSize ;
	message->num++ ;

	addBlock(blocks, offset1, strs1, contig1, size1, exts, rank, storage,status) ;
	addBlock(blocks, offset2, strs2, contig2, size2, exts, rank, storage,status) ;
      }
      else
	newMessage(messages, blocks,
		   offset1, strs1, contig1, size1,
		   offset2, strs2, contig2, size2, 
                   exts, rank, size, storage, status) ;
    }
  }
  else
    splitMessageBlocks(messages, blocks,
		       offset1, strs1, contig1, size1,
		       offset2, strs2, contig2, size2, 
                       exts, rank, status, iclaSize,storage) ;
}



//xh add
static void Col_splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset1, int* strs1, int contig1,int size1,
		              int offset2, int* strs2, int contig2,int size2,
                              int* exts, int rank, Status status, int iclaSize) {

  // Called iff size > STATIC_BUFFER_SIZE.

  // Block is split into a series of messages, with one sub-block
  // per message.

  if(rank == 0)
    {
     newMessage(messages, blocks, 
                offset1, strs1, contig1,size1,
                offset2, strs2, contig2,size2,
                exts, rank, size1+size2, 
                COLUMN_MAJOR, status) ;
        // Single element larger than static buffer size.
        // Don't try to split individual elements---if a buffer is needed,
        // will have to allocate it dynamically at schedule execution time.
//     AdlibPrintf("End splitmessage message=%d\n",messages.size);
    }
  else {
    int subRank = rank - 1 ;
    int count   = exts [subRank] ;
    int stride1  = strs1 [subRank] ;
    int stride2  = strs2 [subRank] ;
    int subSize1 = size1 / count ;
    int subSize2 = size2 / count ;
    int subSize  = subSize1 + subSize2;
    if(subSize <= iclaSize) {

      // Slice leading dimension

      int sliceWidth = iclaSize / subSize ;
      int sliceSize1 = sliceWidth * subSize1 ;
      int sliceSize2 = sliceWidth * subSize2 ;
      int sliceStep1 = sliceWidth * stride1;
      int sliceStep2 = sliceWidth * stride2;
      int numWholeSlices = count / sliceWidth ;

      exts [subRank] = sliceWidth ;

      for(int i = 0 ; i < numWholeSlices ; i++) {
	newMessage(messages, blocks, 
                   offset1, strs1, contig1, sliceSize1,
                   offset2, strs2, contig2, sliceSize2,
                   exts, rank,sliceSize1+sliceSize2,
                   COLUMN_MAJOR, status) ;
	offset1 += sliceStep1 ;
	offset2 += sliceStep2 ;
      }

      int residue = count - numWholeSlices * sliceWidth ;
      if(residue) {
        sliceSize1 = residue * subSize1;
        sliceSize2 = residue * subSize2;

	exts [subRank] = residue ;
	newMessage(messages, blocks, 
                   offset1, strs1, contig1,sliceSize1,
                   offset2, strs2, contig2,sliceSize2,
		   exts, rank,sliceSize1+sliceSize2,
                   COLUMN_MAJOR, status) ;
      }

      exts [subRank] = count ;  // restore to original value.
    }
    else {

      // Reduced rank sub-blocks still too large.  Recurse.

      for(int i = 0  ; i < count ; i++) {
	Col_splitMessageBlock(messages, blocks,
                              offset1, strs1, contig1,subSize1,
                              offset2, strs2, contig2,subSize2,
                              exts, subRank, status, iclaSize) ;
	offset1 += stride1 ;
	offset2 += stride2 ;
      }
    }
  }
}


//xh add
static void Row_splitMessageBlock(MessList &messages, MultiBlockList &blocks,
		              int offset1, int* strs1, int contig1,int size1,
		              int offset2, int* strs2, int contig2,int size2,
                              int* exts, int rank, int d,Status status, int iclaSize) {

  // Called iff size > STATIC_BUFFER_SIZE.

  // Block is split into a series of messages, with one sub-block
  // per message.

  if(rank == 0)
    {
     newMessage(messages, blocks, 
                offset1, strs1, contig1,size1,
                offset2, strs2, contig2,size2,
                exts, rank, size1+size2, 
                ROW_MAJOR, status) ;
        // Single element larger than static buffer size.
        // Don't try to split individual elements---if a buffer is needed,
        // will have to allocate it dynamically at schedule execution time.
//     AdlibPrintf("End splitmessage message=%d\n",messages.size);
    }
  else {
    int subRank = d - 1 ;
    int count   = exts [subRank] ;
    int stride1  = strs1 [subRank] ;
    int stride2  = strs2 [subRank] ;
    int subSize1 = size1 / count ;
    int subSize2 = size2 / count ;
    int subSize  = subSize1 + subSize2;
    if(subSize <= iclaSize) {

      // Slice leading dimension

      int sliceWidth = iclaSize / subSize ;
      int sliceSize1 = sliceWidth * subSize1 ;
      int sliceSize2 = sliceWidth * subSize2 ;
      int sliceStep1 = sliceWidth * stride1;
      int sliceStep2 = sliceWidth * stride2;
      int numWholeSlices = count / sliceWidth ;

      exts [subRank] = sliceWidth ;

      for(int i = 0 ; i < numWholeSlices ; i++) {
	newMessage(messages, blocks, 
                   offset1, strs1+subRank, contig1, sliceSize1,
                   offset2, strs2+subRank, contig2, sliceSize2,
                   exts+subRank, rank-subRank,sliceSize1+sliceSize2,
                   ROW_MAJOR, status) ;
	offset1 += sliceStep1 ;
	offset2 += sliceStep2 ;
      }

      int residue = count - numWholeSlices * sliceWidth ;
      if(residue) {
        sliceSize1 = residue * subSize1;
        sliceSize2 = residue * subSize2;

	exts [subRank] = residue ;
	newMessage(messages, blocks, 
                   offset1, strs1+subRank, contig1,sliceSize1,
                   offset2, strs2+subRank, contig2,sliceSize2,
		   exts+subRank, rank-subRank,sliceSize1+sliceSize2,
                   ROW_MAJOR, status) ;

      }

      exts [subRank] = count ;  // restore to original value.
    }
    else {

      // Reduced rank sub-blocks still too large.  Recurse.

      for(int i = 0  ; i < count ; i++) {
	Row_splitMessageBlock(messages, blocks,
                              offset1, strs1, contig1,subSize1,
                              offset2, strs2, contig2,subSize2,
                              exts, rank, d+1, status, iclaSize) ;
	offset1 += stride1 ;
	offset2 += stride2 ;
      }
    }
  }
}


//xh

static void splitMessageBlocks(MessList &messages, MultiBlockList &blocks,
		              int offset1, int* strs1, int contig1,int size1,
		              int offset2, int* strs2, int contig2,int size2,
                              int* exts, int rank, Status status, 
			      int iclaSize, int storage) 
{
  if (defaultStorageMode == COLUMN_MAJOR || reduceMode == 1)
	Col_splitMessageBlock(messages, blocks, offset1, strs1, contig1,size1,
                              offset2, strs2, contig2, size2,  
                              exts, rank, status, iclaSize);
  else
	Row_splitMessageBlock(messages, blocks, offset1, strs1, contig1,size2,
                              offset2, strs2, contig2, size2,   
                              exts,  rank, 1, status, iclaSize);
}

/*
static void splitMessageBlocks(MessList &messages, MultiBlockList &blocks,
		              int offset1, int* strs1, int contig1, int size1,
		              int offset2, int* strs2, int contig2, int size2,
                              int* exts, int rank) {

  // Called iff size > STATIC_BUFFER_SIZE.

  // Blocks are split into a series of messages, with two sub-blocks
  // per message.

  if(rank == 0) {
    const int a_size1 = ALIGN_UP(size1) ;
    const int a_size2 = ALIGN_UP(size2) ;

    newMessage(messages, blocks,
	       offset1, strs1, contig1, a_size1,
	       offset2, strs2, contig2, a_size2, exts, rank,
               a_size1 + a_size2) ;

        // Single pair of elements larger than static buffer size.
        // Don't try to split individual elements---will have to allocate
        // buffer dynamically at schedule execution time.
  }

  else {
    int subRank = rank - 1 ;
    int count   = exts [subRank] ;
    int stride1 = strs1 [subRank] ;
    int stride2 = strs2 [subRank] ;
    int subSize1 = size1 / count ;
    int subSize2 = size2 / count ;
    int subSize = subSize1 + subSize2 ;
    if(subSize <= STATIC_BUFFER_SIZE - 2 * ALIGN_SZ) {

      // Slice leading dimension

      int sliceWidth = (STATIC_BUFFER_SIZE - 2 * ALIGN_SZ) / subSize ;
      int a_sliceSize1 = ALIGN_UP(sliceWidth * subSize1) ;
      int a_sliceSize2 = ALIGN_UP(sliceWidth * subSize2) ;
      int sliceStep1 = sliceWidth * stride1 ;
      int sliceStep2 = sliceWidth * stride2 ;
      int numWholeSlices = count / sliceWidth ;

      exts [subRank] = sliceWidth ;

      for(int i = 0 ; i < numWholeSlices ; i++) {
	newMessage(messages, blocks,
		   offset1, strs1, contig1, a_sliceSize1,
		   offset2, strs2, contig2, a_sliceSize2,
                   exts, rank, a_sliceSize1 + a_sliceSize2) ;
	offset1 += sliceStep1 ;
	offset2 += sliceStep2 ;
      }

      int residue = count - numWholeSlices * sliceWidth ;
      if(residue) {
	a_sliceSize1 = ALIGN_UP(residue * subSize1) ;
	a_sliceSize2 = ALIGN_UP(residue * subSize2) ;

	exts [subRank] = residue ;
	newMessage(messages, blocks,
                   offset1, strs1, contig1, a_sliceSize1,
                   offset2, strs2, contig2, a_sliceSize2,
                   exts, rank, a_sliceSize1 + a_sliceSize2) ;
      }

      exts [subRank] = count ;  // restore to original value.
    }
    else {

      // Reduced rank sub-blocks still too large.  Recurse.

      for(int i = 0  ; i < count ; i++) {
	splitMessageBlocks(messages, blocks,
                           offset1, strs1, contig1, subSize1,
                           offset2, strs2, contig2, subSize2,
                           exts, subRank) ;
	offset1 += stride1 ;
	offset2 += stride2 ;
      }
    }
  }
}
*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `gatherOffBuf'.  Gather to buffer, given offsets.                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Optimized versions: eliminate `memcpy' calls for some common block sizes...

void gatherOffBuf(void* base,
                  int* offsets, int count, void* buf, int elementLen) {
  switch(elementLen) {
    case 1 : {
      char* cbuf  = (char*) buf  ;
      char* cbase = (char*) base ;
      for(int i = 0 ; i < count ; i++) {
        *cbuf = cbase [*offsets] ;
        cbuf++ ;
        offsets++ ;
      }
      break ;
    }
    case sizeof(int) : {
      int* ibuf  = (int*) buf  ;
      int* ibase = (int*) base ;
      for(int i = 0 ; i < count ; i++) {
        *ibuf = ibase [*offsets] ;
        ibuf++ ;
        offsets++ ;
      }
      break ;
    }
    case sizeof(double) : {
      double* dbuf  = (double*) buf  ;
      double* dbase = (double*) base ;
      for(int i = 0 ; i < count ; i++) {
        *dbuf = dbase [*offsets] ;
        dbuf++ ;
        offsets++ ;
      }
      break ;
    }
    default :
      char* cbuf  = (char*) buf  ;
      char* cbase = (char*) base ;
      for(int i = 0 ; i < count ; i++) {
        memcpy(cbuf, cbase + elementLen * *offsets, elementLen) ;
        cbuf += elementLen ;
        offsets++ ;
      }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `scatterOffBuf'.  Scatter from buffer, given offsets.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void scatterOffBuf(void* base,
                   int* offsets, int count, void* buf, int elementLen) {
  switch(elementLen) {
    case 1 : {
      char* cbuf  = (char*) buf  ;
      char* cbase = (char*) base ;
      for(int i = 0 ; i < count ; i++) {
        cbase [*offsets] = *cbuf ;
        cbuf++ ;
        offsets++ ;
      }
      break ;
    }
    case sizeof(int) : {
      int* ibuf  = (int*) buf  ;
      int* ibase = (int*) base ;
      for(int i = 0 ; i < count ; i++) {
        ibase [*offsets] = *ibuf ;
        ibuf++ ;
        offsets++ ;
      }
      break ;
    }
    case sizeof(double) : {
      double* dbuf  = (double*) buf  ;
      double* dbase = (double*) base ;
      for(int i = 0 ; i < count ; i++) {
        dbase [*offsets] = *dbuf ;
        dbuf++ ;
        offsets++ ;
      }
      break ;
    }
    default :
      char* cbuf  = (char*) buf  ;
      char* cbase = (char*) base ;
      for(int i = 0 ; i < count ; i++) {
        memcpy(cbase + elementLen * *offsets, cbuf, elementLen) ;
        cbuf += elementLen ;
        offsets++ ;
      }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `copyMblkMblk'.  `MultiBlock' to `MultiBlock' copy.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void copyMblkMblk(void* dstBase, MultiBlock dstBlock,
                  void* srcBase, MultiBlock srcBlock, const int elementLen) {

  // `MultiBlock' to `MultiBlock' copy.  The blocks must be the same shape.

  char* dst = (char*) dstBase + elementLen * dstBlock [MBLK_OFF] ;
  char* src = (char*) srcBase + elementLen * srcBlock [MBLK_OFF] ;
  if(dstBlock [MBLK_CTG] && srcBlock [MBLK_CTG])
    memcpy(dst, src, dstBlock [MBLK_SIZ]) ;
  else {

    // (rank > 0, because the scalar case is always contiguous.)

    const int rank      = dstBlock [MBLK_RNK] ;
    int* dstStrs        = dstBlock + MBLK_STR ;
    int* srcStrs        = srcBlock + MBLK_STR ;
    int* exts           = dstStrs  + rank ;

    switch(elementLen) {
      case 1 : {
        blockCopy(dst, dstStrs, src, srcStrs, exts, rank) ;
	break ;
      }
      case sizeof(int) : {
        blockCopy((int*) dst, dstStrs, (int*) src, srcStrs, exts, rank) ;
	break ;
      }
      case sizeof(double) : {
        blockCopy((double*) dst, dstStrs, (double*) src, srcStrs, exts, rank) ;
	break ;
      }
      default :
        blockCopy(dst, dstStrs, src, srcStrs, exts, rank, elementLen) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `gatherMblkBuf'.  Gather to buffer from `MultiBlock'.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gatherMblkBuf(void* base, MultiBlock block, void* buffer,
                   const int elementLen) {

  // `MultiBlock' to buffer' copy.

  char* dat = (char*) base + elementLen * block [MBLK_OFF] ;

  if(block [MBLK_CTG])
    memcpy(buffer, dat, block [MBLK_SIZ]) ;
  else {

    // (rank > 0, because the scalar case is always contiguous.)

    const int rank      = block [MBLK_RNK] ;
    int* strs           = block + MBLK_STR ;
    int* exts           = strs  + rank ;

    int major           = block [MBLK_MAJ];

      // ROW MAJOR transpose the dimension
    if (major == ROW_MAJOR && rank > 1) {
	int *strs0 = strs;
	int *exts0 = exts;
	strs = new int [rank];
	exts = new int [rank];
	for (int i=0; i<rank; i++) {
	   exts[i]=exts0[rank-1-i]; 
	   strs[i]=strs0[rank-1-i]; 
	}
    }

    switch(elementLen) {
      case 1 : {
        blockGather(dat, strs, exts, rank, (char*) buffer) ;
	break ;
      }
      case sizeof(int) : {
        blockGather((int*) dat, strs, exts, rank, (int*) buffer) ;
	break ;
      }
      case sizeof(double) : {
        blockGather((double*) dat, strs, exts, rank, (double*) buffer) ;
	break ;
      }
      default :
        blockGather(dat, strs, exts, rank, (char*) buffer, elementLen) ;
    }

    if (major == ROW_MAJOR && rank > 1) {
	delete strs;
	delete exts;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `scatterMblkBuf'.  Scatter from buffer to `MultiBlock'.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void scatterMblkBuf(void* base, MultiBlock block, void* buffer,
                    const int elementLen) {

  // `MultiBlock' to buffer' copy.

  char* dat = (char*) base + elementLen * block [MBLK_OFF] ;
  if(block [MBLK_CTG])
    memcpy(dat, buffer, block [MBLK_SIZ]) ;
  else {
    // (rank > 0, because the scalar case is always contiguous.)

    const int rank      = block [MBLK_RNK] ;
    int* strs           = block + MBLK_STR ;
    int* exts           = strs  + rank ;

    int major           = block [MBLK_MAJ];

      // ROW MAJOR transpose the dimension
    if (major == ROW_MAJOR && rank > 1) {
	int *strs0 = strs;
	int *exts0 = exts;
	strs = new int [rank];
	exts = new int [rank];
	for (int i=0; i<rank; i++) {
	   exts[i]=exts0[rank-1-i]; 
	   strs[i]=strs0[rank-1-i]; 
	}
    }

    switch(elementLen) {
      case 1 : {
        blockScatter(dat, strs, exts, rank, (char*) buffer) ;
	break ;
      }
      case sizeof(int) : {
        blockScatter((int*) dat, strs, exts, rank, (int*) buffer) ;
	break ;
      }
      case sizeof(double) : {
        blockScatter((double*) dat, strs, exts, rank, (double*) buffer) ;
	break ;
      }
      default :
        blockScatter(dat, strs, exts, rank, (char*) buffer, elementLen) ;
    }

    if (major == ROW_MAJOR && rank > 1) {
	delete strs;
	delete exts;
    }
  }

}


// -------------------------------------------------------------
// Things to do
//
//   Check for step == 1 in rank == 1 cases of block gather/scatter/copy
//   routines?  Inline code for rank == 1 in top level routines?
//
//   Have `copyMblkMblk' invoke `blockGather' or `blockScatter' if *one*
//   block is contiguous.
//
//   Eliminate return value of `blockGather'...  Make `buffer' argument
//   a reference?
//
//   Probably more practical to literally put `newMessage' inline.
//
//   Do something about `addMessageBlocks'.  It's too much of a hack.

