//
//	ProcInfo.cc
//		Implement performance data classes
//		definition in ProcInfo.h
//

#define _CPP
#define _PROCINFO
#define _RUNTIME
#include "headers.h"

/////////////////////////////////////////////////////////////////////////
//	CProcInfo:	container class of all performance data of a processor
int CProcInfo::ms_nVersion=1;

CProcInfo::CProcInfo()
{
	m_nProcID=0;
	m_nTimeCPU=0;
	m_nTimeSystem=0;

	m_nElapseCPU=0;
	m_nElapseSystem=0;

	for (int i=0; i<MAX_PROC; i++)
	{
		m_aSend[i]=0;	
		m_aSendAmount[i]=0;	
		m_aRecv[i]=0;	
		m_aRecvAmount[i]=0;	
	}
}

BOOL CProcInfo::Start(int nProcID)
{
	m_nTimeCPU=GetTimeCPU();
	m_nTimeSystem=GetTimeSystem();
	m_nProcID=nProcID;

	m_nElapseCPU=0;
	m_nElapseSystem=0;

	for (int i=0; i<MAX_PROC; i++)
	{
		m_aSend[i]=0;
		m_aSendAmount[i]=0;
		m_aRecv[i]=0;
		m_aRecvAmount[i]=0;
	}

	while (!m_aStmts.empty())
		m_aStmts.pop_back();
	while (!m_aEvents.empty())
		m_aEvents.pop_back();
	while (!m_aDistArrays.empty())
		m_aDistArrays.pop_back();

	NewEvent(0, stat_serial);
	return TRUE;
}

BOOL CProcInfo::Stop()
{
	m_nElapseCPU=GetTimeCPU()-m_nTimeCPU;
	m_nElapseSystem=GetTimeSystem()-m_nTimeSystem;
	NewEvent(0, stat_term);
	Write(stderr);
	return TRUE;
}

BOOL CProcInfo::GetSendTotal(int &nCount, int &nAmount)
{
	nCount=0;
	nAmount=0;
	for (int i=0; i<MAX_PROC; i++)
	{
		nCount+=m_aSend[i];
		nAmount+=m_aSendAmount[i];
	}
	return TRUE;
}

BOOL CProcInfo::GetRecvTotal(int &nCount, int &nAmount)
{
	nCount=0;
	nAmount=0;
	for (int i=0; i<MAX_PROC; i++)
	{
		nCount+=m_aRecv[i];
		nAmount+=m_aRecvAmount[i];
	}
	return TRUE;
}

CParaStmtInfo *CProcInfo::NewStmt(int nStmtID)
{
	CParaStmtInfo stmt(nStmtID);
	m_aStmts.insert(m_aStmts.end(), stmt);
	return &(m_aStmts.back());
}

CParaStmtInfo *CProcInfo::FindStmt(int nStmtID)
{
	for (int i=0; i<m_aStmts.size(); i++)
	{
		if (m_aStmts[i].m_nStmtID==nStmtID)
			return &(m_aStmts[i]);
	}
	return NULL;
}

POSITION CProcInfo::GetFirstStmtPos()
{
	//	ignore stmt id for now
	if (m_aStmts.size()<=0)
		return -1;
	return 0;
}

POSITION CProcInfo::GetLastStmtPos()
{
	//	ignore stmt id for now
	if (m_aStmts.size()<=0)
		return -1;
	return m_aStmts.size()-1;
}

CParaStmtInfo *CProcInfo::GetStmt(POSITION pos)
{
	if (pos<0 || pos>=m_aStmts.size())
		return NULL;

	return &(m_aStmts[pos]);
}

CParaStmtInfo *CProcInfo::GetNextStmt(POSITION &pos)
{
	if (pos<0 || pos>=m_aStmts.size())
		return NULL;

	pos++;
	return &(m_aStmts[pos-1]);
}

CParaStmtInfo *CProcInfo::GetPrevStmt(POSITION &pos)
{
	if (pos<0 || pos>=m_aStmts.size())
		return NULL;

	pos--;
	return &(m_aStmts[pos+1]);
}

CExecEvent *CProcInfo::NewEvent(int nStmtID, int nState, 
		int nComTarget, int nComSize)
{
	if (nState==stat_para)
		CExecEvent::LevelIn();
	else if (nState==stat_endpara)
		CExecEvent::LevelOut();

	CExecEvent event(nStmtID);
	event.m_nState=nState;
	event.m_nComTarget=nComTarget;
	event.m_nComSize=nComSize;
	m_aEvents.insert(m_aEvents.end(), event);
	return &(m_aEvents.back());
}

BOOL CProcInfo::EraseLastEvent()
{
	if (m_aEvents.size()>0)
		m_aEvents.pop_back();
	return TRUE;
}

POSITION CProcInfo::GetFirstEventPos(int nStmtID)
{
	//	ignore stmt id for now
	if (m_aEvents.size()<=0)
		return -1;
	return 0;
}

POSITION CProcInfo::GetLastEventPos(int nStmtID)
{
	//	ignore stmt id for now
	if (m_aEvents.size()<=0)
		return -1;
	return m_aEvents.size()-1;
}

CExecEvent *CProcInfo::GetEvent(POSITION pos)
{
	if (pos<0 || pos>=m_aEvents.size())
		return NULL;

	return &(m_aEvents[pos]);
}

CExecEvent *CProcInfo::GetNextEvent(POSITION &pos, int nStmtID)
{
	if (pos<0 || pos>=m_aEvents.size())
		return NULL;

	pos++;
	return &(m_aEvents[pos-1]);
}

CExecEvent *CProcInfo::GetPrevEvent(POSITION &pos, int nStmtID)
{
	if (pos<0 || pos>=m_aEvents.size())
		return NULL;

	pos--;
	return &(m_aEvents[pos+1]);
}
	
POSITION CProcInfo::FindEventByTime(int nTimeCPU)
{
	int nDown=0;
	int nUp=m_aEvents.size()-1;
	for (int i=m_aEvents.size()/2; i>=0 && i<m_aEvents.size(); i=(nUp+nDown)/2)
	{
		if (m_aEvents[i].m_nTimeCPU>=nTimeCPU)
		{
			nUp=i-1;
			if (nDown>nUp)
			{
				return i;
			}
		}
		else
		{
			nDown=i+1;
			if (nDown>nUp)
			{
				return -1;
			}
		}
	}
	return -1;
}

POSITION CProcInfo::FindEventBySysTime(int nTimeSystem)
{
	int nDown=0;
	int nUp=m_aEvents.size()-1;
	for (int i=m_aEvents.size()/2; i>=0 && i<m_aEvents.size(); i=(nUp+nDown)/2)
	{
		if (m_aEvents[i].m_nTimeSystem>=nTimeSystem)
		{
			nUp=i-1;
			if (nDown>nUp)
			{
				return i;
			}
		}
		else
		{
			nDown=i+1;
			if (nDown>nUp)
			{
				return nDown;
			}
		}
	}
	return -1;
}

POSITION CProcInfo::FindEventByTime(int nTimeCPU, int nSpan, POSITION &posNext)
{
	int nDown=0;
	int nUp=m_aEvents.size()-1;
	for (int i=m_aEvents.size()/2; i>=0 && i<m_aEvents.size(); i=(nUp+nDown)/2)
	{
		if (m_aEvents[i].m_nTimeCPU<=nTimeCPU+nSpan &&
			m_aEvents[i].m_nElapseCPU+m_aEvents[i].m_nTimeCPU>nTimeCPU)
		{	//	found
			int nCenter=i;

			while ( m_aEvents[i].m_nTimeCPU<=nTimeCPU+nSpan && i<=nUp)
			{
				if (m_aEvents[i].m_nState==stat_send ||
					m_aEvents[i].m_nState==stat_recv)
				{	//	found
					posNext=i;
					return i;
				}
				i++;
			}
			i=nCenter;
			while (m_aEvents[i].m_nElapseCPU+m_aEvents[i].m_nTimeCPU>nTimeCPU &&
				i>=nDown)
			{
				i--;
				if (m_aEvents[i].m_nState==stat_send ||
					m_aEvents[i].m_nState==stat_recv)
				{	//	found
					posNext=i;
					return i;
				}
			}
			//	not found
			posNext=nCenter;
			return -1;
		}
		else if (m_aEvents[i].m_nTimeCPU>nTimeCPU+nSpan)
		{
			nUp=i-1;
			if (nDown>nUp)
			{
				posNext=i;
				return -1;
			}
		}
		else
		{
			nDown=i+1;
			if (nDown>nUp)
			{
				posNext=i+1;
				return -1;
			}
		}
	}
	return -1;
}

POSITION CProcInfo::FindEventBySysTime(int nTimeSystem, 
	int nSpan, POSITION &posNext)
{
	int nDown=0;
	int nUp=m_aEvents.size()-1;
	for (int i=m_aEvents.size()/2; i>=0 && i<m_aEvents.size(); i=(nUp+nDown)/2)
	{
		if (m_aEvents[i].m_nTimeSystem<=nTimeSystem+nSpan &&
			m_aEvents[i].m_nElapseSystem+m_aEvents[i].m_nTimeSystem>nTimeSystem)
		{	//	found
			int nCenter=i;

			while ( m_aEvents[i].m_nTimeSystem<=nTimeSystem+nSpan && i<=nUp)
			{
				if (m_aEvents[i].m_nState==stat_send ||
					m_aEvents[i].m_nState==stat_recv)
				{	//	found
					posNext=i;
					return i;
				}
				i++;
			}
			i=nCenter-1;
			while (m_aEvents[i].m_nElapseSystem+m_aEvents[i].m_nTimeSystem
				>nTimeSystem &&
				i>=nDown)
			{
				if (m_aEvents[i].m_nState==stat_send ||
					m_aEvents[i].m_nState==stat_recv)
				{	//	found
					posNext=i;
					return i;
				}
				i--;
			}
			//	not found
			posNext=nCenter;
			return -1;
		}
		else if (m_aEvents[i].m_nTimeSystem>nTimeSystem+nSpan)
		{
			nUp=i-1;
			if (nDown>nUp)
			{
				posNext=i;
				return -1;
			}
		}
		else
		{
			nDown=i+1;
			if (nDown>nUp)
			{
				posNext=i+1;
				return -1;
			}
		}
	}
	return -1;
}

BOOL CProcInfo::GetStatus(int &nStatus, POSITION pos, POSITION posNext)
{
	if (pos==-1 && (posNext==-1 || posNext>=m_aEvents.size()))
	{
		nStatus=stat_none;
		return FALSE;
	}
	if (pos==-1)
	{
		CExecEvent *pEvent=&m_aEvents[posNext];
		if (pEvent==NULL)
		{
			nStatus=stat_none;
			return FALSE;
		}
		switch (pEvent->m_nState)
		{
		case stat_para:
			if (pEvent->m_nLevel>1)
				nStatus=stat_para;
			else nStatus=stat_serial;
			break;
	
		case stat_endpara:
			nStatus=stat_para;
			break;
	
		case stat_send:
		case stat_recv:
			if (pEvent->m_nLevel>0)
				nStatus=stat_para;
			else nStatus=stat_serial;
			break;

		case stat_serial:
		case stat_term:
			nStatus=stat_serial;
			break;

		default:
			nStatus=stat_none;
		}
	}
	else
	{
		CExecEvent *pEvent=&m_aEvents[pos];
		switch (pEvent->m_nState)
		{
		case stat_para:
			if (pEvent->m_nLevel>1)
				nStatus=stat_para;
			else nStatus=stat_serial;
			break;
	
		case stat_endpara:
			nStatus=stat_para;
			break;
	
		case stat_send:
		case stat_recv:
			nStatus=pEvent->m_nState;
			break;

		default:
			nStatus=stat_none;
		}
	}
	return TRUE;
}

BOOL CProcInfo::GetStatusByTime(int nTimeCPU, int nSpan,
	int &nStatus, POSITION &pos)
{
	POSITION posNext;

	pos=FindEventByTime(nTimeCPU, nSpan, posNext);
	return GetStatus(nStatus, pos, posNext);
}

BOOL CProcInfo::GetStatusBySysTime(int nTimeSystem, int nSpan,
	int &nStatus, POSITION &pos)
{
	POSITION posNext;

	pos=FindEventBySysTime(nTimeSystem, nSpan, posNext);
	return GetStatus(nStatus, pos, posNext);
}

BOOL CProcInfo::GetCommByTime(int nTimeCPU, int nSpan, 
	int aSend[MAX_PROC], int aSendAmount[MAX_PROC])
{
	POSITION pos1=FindEventByTime(nTimeCPU);
	POSITION pos2=FindEventByTime(nTimeCPU+nSpan);

	for (int pos=pos1; pos<pos2; pos++)
	{
		if (m_aEvents[pos].m_nState==stat_send)
		{
			aSend[m_aEvents[pos].m_nComTarget]++;
			aSendAmount[m_aEvents[pos].m_nComTarget]+=m_aEvents[pos].m_nComSize;
		}
	}
	return TRUE;
}

BOOL CProcInfo::GetCommBySysTime(int nTimeSystem, int nSpan,
	int *aSend, int *aSendAmount)
{
	POSITION pos1=FindEventBySysTime(nTimeSystem);
	POSITION pos2=FindEventBySysTime(nTimeSystem+nSpan);

	for (int i=0; i<MAX_PROC; i++)
	{
		aSend[i]=aSendAmount[i]=0;
	}
	for (int pos=pos1; pos<pos2; pos++)
	{
		if (m_aEvents[pos].m_nState==stat_send &&
			m_aEvents[pos].m_nComTarget>=0 &&
			m_aEvents[pos].m_nComTarget<MAX_PROC)
		{
			aSend[m_aEvents[pos].m_nComTarget]++;
			aSendAmount[m_aEvents[pos].m_nComTarget]+=m_aEvents[pos].m_nComSize;
		}
	}
	return TRUE;
}

CProcessors *CProcInfo::NewProcDef(int nProcessorsID, int nGroupID)
{
	CProcessors procs;

	procs.m_nProcessorsID=nProcessorsID;
	procs.m_nGrp=nGroupID;
	m_aProcDefs.insert(m_aProcDefs.end(), procs);
	return &(m_aProcDefs.back());
}

CProcessors *CProcInfo::FindProcDef(int nProcessorsID)
{
	for (int i=0; i<m_aProcDefs.size(); i++)
	{
		if (m_aProcDefs[i].m_nProcessorsID==nProcessorsID)
			return &(m_aProcDefs[i]);
	}
	return NULL;
}

CProcessors *CProcInfo::FindProcDefByGrp(int nGroupID)
{
	for (int i=0; i<m_aProcDefs.size(); i++)
	{
		if (m_aProcDefs[i].m_nGrp==nGroupID)
			return &(m_aProcDefs[i]);
	}
	return NULL;
}

CTemplate *CProcInfo::NewTemplate(int nTemplateID, int nRangeID, int nDim)
{
	CTemplate templ;

	templ.m_nTemplateID=nTemplateID;
	templ.m_nDim=nDim;
	templ.m_nRangeID=nRangeID;
	m_aTemplates.insert(m_aTemplates.end(), templ);
	return &(m_aTemplates.back());
}

CTemplate *CProcInfo::FindTemplate(int nTemplateID, int nDim)
{
	for (int i=0; i<m_aTemplates.size(); i++)
	{
		if (m_aTemplates[i].m_nTemplateID==nTemplateID &&
			m_aTemplates[i].m_nDim==nDim)
			return &(m_aTemplates[i]);
	}
	return NULL;
}

CTemplate *CProcInfo::FindTemplateByRange(int nRangeID)
{
	for (int i=0; i<m_aTemplates.size(); i++)
	{
		if (m_aTemplates[i].m_nRangeID==nRangeID)
			return &(m_aTemplates[i]);
	}
	return NULL;
}

CDistArrayInfo *CProcInfo::NewDistArray(int nArrayID, int nDAD)
{
	CDistArrayInfo distArray;
	
	distArray.m_nArrayID=nArrayID;
	distArray.m_nDAD=nDAD;
	m_aDistArrays.insert(m_aDistArrays.end(), distArray);
	return &(m_aDistArrays.back());
}

CDistArrayInfo *CProcInfo::FindDistArray(int nArrayID)
{
	for (int i=0; i<m_aDistArrays.size(); i++)
	{
		if (m_aDistArrays[i].m_nArrayID==nArrayID)
			return &(m_aDistArrays[i]);
	}
	return NULL;
}

CDistArrayInfo *CProcInfo::FindDistArrayByDAD(int nDAD)
{
	for (int i=0; i<m_aDistArrays.size(); i++)
	{
		if (m_aDistArrays[i].m_nDAD==nDAD)
			return &(m_aDistArrays[i]);
	}
	return NULL;
}

POSITION CProcInfo::GetFirstArrayPos()
{
	//	ignore stmt id for now
	if (m_aDistArrays.size()<=0)
		return -1;
	return 0;
}

POSITION CProcInfo::GetLastArrayPos()
{
	//	ignore stmt id for now
	if (m_aDistArrays.size()<=0)
		return -1;
	return m_aDistArrays.size()-1;
}

CDistArrayInfo *CProcInfo::GetArray(POSITION pos)
{
	if (pos<0 || pos>=m_aDistArrays.size())
		return NULL;

	return &(m_aDistArrays[pos]);
}

CDistArrayInfo *CProcInfo::GetNextArray(POSITION &pos)
{
	if (pos<0 || pos>=m_aDistArrays.size())
		return NULL;

	pos++;
	return &(m_aDistArrays[pos-1]);
}

CDistArrayInfo *CProcInfo::GetPrevArray(POSITION &pos)
{
	if (pos<0 || pos>=m_aDistArrays.size())

	pos--;
	return &(m_aDistArrays[pos+1]);
}

/////////////////////////////////////////////////////////////////////////
//	CParaStmtInfo:	class of statistical infomation of a parallel statement
int CParaStmtInfo::ms_nVersion=1;

CParaStmtInfo::CParaStmtInfo()
{
	m_nStmtID=0;
	m_nLoopCounter=0;		
	m_nElapseCPU=0;		
	m_nElapseSystem=0;	
	m_nTotalElapseCPU=0;	
	m_nTotalElapseSystem=0;	
	m_nComElapseCPU=0;	
	m_nComElapseSystem=0;	
	m_nShift=0;			
	m_nShiftAmount=0;		
	m_nRemap=0;			
	m_nRemapAmount=0;		
	for (int i=0; i<MAX_PROC; i++)
	{
		m_aSend[i]=0;	
		m_aSendAmount[i]=0;	
		m_aRecv[i]=0;	
		m_aRecvAmount[i]=0;	
	}
	m_nSend=0;	
	m_nSendAmount=0;	
	m_nRecv=0;	
	m_nRecvAmount=0;	
}

CParaStmtInfo::CParaStmtInfo(int nStmtID)
{
	m_nStmtID=nStmtID;
	m_nLoopCounter=0;
	m_nElapseCPU=0;
	m_nElapseSystem=0;
	m_nTotalElapseCPU=0;
	m_nTotalElapseSystem=0;
	m_nComElapseCPU=0;
	m_nComElapseSystem=0;
	m_nShift=0;
	m_nShiftAmount=0;
	m_nRemap=0;
	m_nRemapAmount=0;
	for (int i=0; i<MAX_PROC; i++)
	{
	    m_aSend[i]=0;
	    m_aSendAmount[i]=0;
	    m_aRecv[i]=0;
	    m_aRecvAmount[i]=0;
	}
	m_nSend=0;	
	m_nSendAmount=0;	
	m_nRecv=0;	
	m_nRecvAmount=0;	
}

CParaStmtInfo &CParaStmtInfo::operator =(const CParaStmtInfo &stmt)
{
	m_nStmtID=stmt.m_nStmtID;
	m_nLoopCounter=stmt.m_nLoopCounter;		
	m_nElapseCPU=stmt.m_nElapseCPU;		
	m_nElapseSystem=stmt.m_nElapseSystem;	
	m_nTotalElapseCPU=stmt.m_nTotalElapseCPU;	
	m_nTotalElapseSystem=stmt.m_nTotalElapseSystem;	
	m_nComElapseCPU=stmt.m_nComElapseCPU;	
	m_nComElapseSystem=stmt.m_nComElapseSystem;	
	m_nShift=stmt.m_nShift;			
	m_nShiftAmount=stmt.m_nShiftAmount;		
	m_nRemap=stmt.m_nRemap;			
	m_nRemapAmount=stmt.m_nRemapAmount;		
	for (int i=0; i<MAX_PROC; i++)
	{
		m_aSend[i]=stmt.m_aSend[i];	
		m_aSendAmount[i]=stmt.m_aSendAmount[i];	
		m_aRecv[i]=stmt.m_aRecv[i];	
		m_aRecvAmount[i]=stmt.m_aRecvAmount[i];	
	}
	m_nSend=stmt.m_nSend;	
	m_nSendAmount=stmt.m_nSendAmount;	
	m_nRecv=stmt.m_nRecv;	
	m_nRecvAmount=stmt.m_nRecvAmount;	

	return *this;
}

/////////////////////////////////////////////////////////////////////////
//	CExecEvent:	class of event in execution
int CExecEvent::ms_nVersion=2;
int CExecEvent::ms_nNextID=1;
int CExecEvent::ms_nLevel=0;

CExecEvent::CExecEvent()
{
	m_nID=NewID();
	m_nStmtID=0;		
	m_nState=0;		
	m_nTimeCPU=GetTimeCPU();		
	m_nTimeSystem=GetTimeSystem();	
	m_nElapseCPU=0;		
	m_nElapseSystem=0;	

	m_nComTarget=0;	
	m_nComSize=0;		
	m_nRepCount=0;	
						
	m_nEnclosingID=0;	

	m_nLevel=ms_nLevel;
}

CExecEvent::CExecEvent(int nStmtID)
{
	m_nID=NewID();
	m_nStmtID=nStmtID;
	m_nState=0;
	m_nTimeCPU=GetTimeCPU();
	m_nTimeSystem=GetTimeSystem();
	m_nElapseCPU=0;
	m_nElapseSystem=0;

	m_nComTarget=0;
	m_nComSize=0;
	m_nRepCount=0;
	   
	m_nEnclosingID=0;

	m_nLevel=ms_nLevel;
}

CExecEvent &CExecEvent::operator =(const CExecEvent &event)
{
	m_nID=event.m_nID;
	m_nStmtID=event.m_nStmtID;		
	m_nState=event.m_nState;		
	m_nTimeCPU=event.m_nTimeCPU;
	m_nTimeSystem=event.m_nTimeSystem;
	m_nElapseCPU=event.m_nElapseCPU;		
	m_nElapseSystem=event.m_nElapseSystem;	
	m_nComTarget=event.m_nComTarget;	
	m_nComSize=event.m_nComSize;		
	m_nRepCount=event.m_nRepCount;	
	m_nEnclosingID=event.m_nEnclosingID;	
	m_nLevel=event.m_nLevel;

	return *this;
}

BOOL CExecEvent::EndEvent()
{
	m_nElapseCPU=GetTimeCPU()-m_nTimeCPU;
	m_nElapseSystem=GetTimeSystem()-m_nTimeSystem;
	
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
//	data distribution classes
int CProcessors::ms_nVersion=1;

CProcessors::CProcessors()
{
	m_nProcessorsID=0;
	m_nGrp=0;
	m_nDimCount=0;
	for (int i=0; i<MAX_LEVEL; i++)
		m_aSize[i]=0;
}

CProcessors &CProcessors::operator =(const CProcessors &procs)
{
	m_nProcessorsID=procs.m_nProcessorsID;
	m_nGrp=procs.m_nGrp;
	m_nDimCount=procs.m_nDimCount;
	for (int i=0; i<MAX_LEVEL; i++)
		m_aSize[i]=procs.m_aSize[i];
	return *this;
}


int CTemplate::ms_nVersion=1;

CTemplate::CTemplate()
{
	m_nTemplateID=0;
	m_nDim=0;
	m_nRangeID=0;
	m_nLB=0;
	m_nUB=0;
	m_nDistGroup=0;
	m_nDistType=dist_copy;
	m_nBlockSize=0;
}

CTemplate &CTemplate::operator =(const CTemplate &templ)
{
	m_nTemplateID=templ.m_nTemplateID;
	m_nDim=templ.m_nDim;
	m_nRangeID=templ.m_nRangeID;
	m_nLB=templ.m_nLB;
	m_nUB=templ.m_nUB;
	m_nDistGroup=templ.m_nDistGroup;
	m_nDistType=templ.m_nDistType;
	m_nBlockSize=templ.m_nBlockSize;
	return *this;
}

CTemplate::SetDist(int lb, int ub, int nGroup, int nType, int nBlockSize)
{
	m_nLB=lb;
	m_nUB=ub;
	m_nDistGroup=nGroup;
	m_nDistType=nType;
	m_nBlockSize=nBlockSize;
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
//	CDistArrayInfo:	class of distributed array 
int CDistArrayInfo::ms_nVersion=1;

CDistArrayInfo::CDistArrayInfo()
{
	m_nArrayID=0;
	m_nDAD=0;
	m_nDimAmount=0;
	
	for (int i=0; i<MAX_LEVEL; i++)
	{
		m_aLB[i]=0;
		m_aUB[i]=0;
		m_aAlignRange[i]=0;
		m_aAlignType[i]=align_copy;
		m_a[i]=0; 
		m_b[i]=0;
	}
}

CDistArrayInfo &CDistArrayInfo::operator =(const CDistArrayInfo &array)
{
	m_nArrayID=array.m_nArrayID;
	m_nDAD=array.m_nDAD;
	m_nDimAmount=array.m_nDimAmount;
	
	for (int i=0; i<MAX_LEVEL; i++)
	{
		m_aLB[i]=array.m_aLB[i];
		m_aUB[i]=array.m_aUB[i];
		m_aAlignRange[i]=array.m_aAlignRange[i];
		m_aAlignType[i]=array.m_aAlignType[i];
		m_a[i]=array.m_a[i]; 
		m_b[i]=array.m_b[i];
	}
	return *this;
}

BOOL CDistArrayInfo::SetArrayAlign(int nDim, int nLB, int nUB,
		int a, int b, int nRange, int nType)
{
	if (nDim<=0 || nDim>MAX_LEVEL) return FALSE;
	nDim--;
	m_aLB[nDim]=nLB;
	m_aUB[nDim]=nUB;
	m_aAlignRange[nDim]=nRange;
	m_aAlignType[nDim]=nType;
	m_a[nDim]=a; 
	m_b[nDim]=b;
	return TRUE;
}
	
////////////////////////////////////////////////////////////////////////////
//	data structure I/O code
#define KW_PROC "Proc"
#define KW_COMM "Comm"
#define KW_ENDPROC "EndProc"
#define KW_VERSION "Ver"
#define KW_STMT "Stmt"
#define KW_ENDSTMT "EndStmt"
#define KW_EVENT "Event"
#define KW_PROCESSORS "Processors"
#define KW_TEMPLATE "Template"
#define KW_ARRAY "Array"

#define CPU2usec(clk) ( clk * ( 1000000 / CLOCKS_PER_SEC) )
//#define CPU2usec(clk) clk

static char szBuf[1024];
static char szSymbol[1024];
static int nLineCount;
static int nProcID;;

BOOL CProcInfo::Read(FILE *fp)
{
	while (!m_aStmts.empty())
		m_aStmts.pop_back();
	while (!m_aEvents.empty())
		m_aEvents.pop_back();
	while (!m_aDistArrays.empty())
		m_aDistArrays.pop_back();

	BOOL bMatchHeader=FALSE;
	int nVersion;

	while (!bMatchHeader)
	{
		if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp))
			return FALSE;
		sscanf(szBuf, "%d %d %s", &nProcID, &nLineCount, szSymbol);
		if (!strcmp(KW_PROC, szSymbol))
			bMatchHeader=TRUE;
	}

	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d", 
		&nProcID, &nLineCount, &m_nProcID, szSymbol, &nVersion);
	
	if (nVersion==1)
	{
		//	parse file of version 1
		int nProc=0;

		if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
		sscanf(szBuf, "%d %d %d %d", 
			&nProcID, &nLineCount, &m_nElapseCPU, &m_nElapseSystem);
		while (NULL!=fgets(szBuf, sizeof(szBuf)-1, fp))
		{
			sscanf(szBuf, "%d %d %s", &nProcID, &nLineCount, szSymbol);
			if (!strcmp(szSymbol, KW_COMM))
			{
				sscanf(szBuf, "%d %d %s %d %d %d %d", &nProcID, &nLineCount,
					szSymbol, &m_aSend[nProc], &m_aSendAmount[nProc],
					&m_aRecv[nProc], &m_aRecvAmount[nProc]);
				nProc++;
			}
			else if (!strcmp(szSymbol, KW_STMT))
			{
				CParaStmtInfo *pStmt=NewStmt(0);
				if (pStmt!=NULL)
					pStmt->Read(fp);
			}
			else if (!strcmp(szSymbol, KW_PROCESSORS))
			{
				CProcessors *pProcs=NewProcDef(0,0);
				if (pProcs!=NULL)
					pProcs->Read(fp);
			}
			else if (!strcmp(szSymbol, KW_TEMPLATE))
			{
				CTemplate *pTempl=NewTemplate(0,0,0);
				if (pTempl!=NULL)
					pTempl->Read(fp);
			}
			else if (!strcmp(szSymbol, KW_ARRAY))
			{
				CDistArrayInfo *pArray=NewDistArray(0,0);
				if (pArray!=NULL)
					pArray->Read(fp);
			}
			else if (!strcmp(szSymbol, KW_EVENT))
			{
				CExecEvent *pEvent=NewEvent(0, 0);
				if (pEvent!=NULL)
					pEvent->Read(fp);
			}
			else if (!strcmp(szSymbol, KW_ENDPROC))
			{
				break;
			}
		}
	}
	return TRUE;
}

BOOL CProcInfo::Write(FILE *fp)
{
	nLineCount=0;
	nProcID=m_nProcID;

	//	format of version 1:
	//
	//	Proc
	//		id Ver ver
	//		elapseCPU elapseSystem
	//		Comm s0 sa0 r0 ra0
	//		Comm s1 sa1 r1 ra1
	//		Comm s2 sa2 r2 ra2
	//		...
	//		Stmt
	//			one stmt
	//		Stmt
	//			one stmt
	//		...
	//		Event
	//			one event
	//		Event
	//			one event
	//		...
	//		Array
	//			one array
	//		...
	//	EndProc
	//

	sprintf(szBuf, "%03d %06d %s\n", nProcID, nLineCount++, KW_PROC);
	fprintf(fp, "%s", szBuf);
	sprintf(szBuf, "%03d %06d \t%d %s %d\n", nProcID, nLineCount++,
		m_nProcID, KW_VERSION, ms_nVersion);
	fprintf(fp, "%s", szBuf);
	sprintf(szBuf, "%03d %06d \t%d %d\n", nProcID, nLineCount++,
		CPU2usec(m_nElapseCPU), m_nElapseSystem);
	fprintf(fp, "%s", szBuf);

	for (int i=0; i<GetMaxProc(); i++)
	{
		sprintf(szBuf, "%03d %06d \t\t%s %d %d %d %d\n", nProcID, nLineCount++, 
			KW_COMM, m_aSend[i], m_aSendAmount[i], 
			m_aRecv[i], m_aRecvAmount[i]);
		fprintf(fp, "%s", szBuf);
	}

	for (int i=0; i<m_aStmts.size(); i++)
	{
		sprintf(szBuf, "%03d %06d \t%s\n", nProcID, nLineCount++, KW_STMT);
		fprintf(fp, "%s", szBuf);
		m_aStmts[i].Write(fp);
	}
	for (int i=0; i<m_aEvents.size(); i++)
	{
		sprintf(szBuf, "%03d %06d \t%s\n", nProcID, nLineCount++, KW_EVENT);
		fprintf(fp, "%s", szBuf);
		m_aEvents[i].Write(fp);
	}
	for (int i=0; i<m_aProcDefs.size(); i++)
	{
		sprintf(szBuf, "%03d %06d \t%s\n", 
			nProcID, nLineCount++, KW_PROCESSORS);
		fprintf(fp, "%s", szBuf);
		m_aProcDefs[i].Write(fp);
	}
	for (int i=0; i<m_aTemplates.size(); i++)
	{
		sprintf(szBuf, "%03d %06d \t%s\n", nProcID, nLineCount++, KW_TEMPLATE);
		fprintf(fp, "%s", szBuf);
		m_aTemplates[i].Write(fp);
	}
	for (int i=0; i<m_aDistArrays.size(); i++)
	{
		sprintf(szBuf, "%03d %06d \t%s\n", nProcID, nLineCount++, KW_ARRAY);
		fprintf(fp, "%s", szBuf);
		m_aDistArrays[i].Write(fp);
	}
	sprintf(szBuf, "%03d %06d %s\n", nProcID, nLineCount++, KW_ENDPROC);
	fprintf(fp, "%s", szBuf);

	return TRUE;
}

BOOL CParaStmtInfo::Read(FILE *fp)
{
	int nVersion;
	int nProcCount=0;

	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d", 
		&nProcID, &nLineCount, &m_nStmtID, szSymbol, &nVersion);
	
	if (nVersion==1)
	{
		if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
		sscanf(szBuf, "%d %d %d %d %d %d %d %d %d %d %d %d %d", 
			&nProcID, &nLineCount,
			&m_nLoopCounter,     
			&m_nElapseCPU,       
			&m_nElapseSystem,    
			&m_nTotalElapseCPU,  
			&m_nTotalElapseSystem,   
			&m_nComElapseCPU,    
			&m_nComElapseSystem, 
			&m_nShift,           
			&m_nShiftAmount,     
			&m_nRemap,           
			&m_nRemapAmount);     
		
		while (NULL!=fgets(szBuf, sizeof(szBuf)-1, fp))
		{
			sscanf(szBuf, "%d %d %s", &nProcID, &nLineCount, szSymbol);
			if (!strcmp(szSymbol, KW_ENDSTMT))
			{
				break;
			}
			else
			{
				if (nProcCount<MAX_PROC)
				{
					sscanf(szBuf, "%d %d %d %d %d %d", &nProcID, &nLineCount,
						&m_aSend[nProcCount], &m_aSendAmount[nProcCount], 
						&m_aRecv[nProcCount], &m_aRecvAmount[nProcCount]);
					m_nSend+=m_aSend[nProcCount];
					m_nRecv+=m_aRecv[nProcCount];
					m_nSendAmount+=m_aSendAmount[nProcCount];
					m_nRecvAmount+=m_aRecvAmount[nProcCount];
				}
				nProcCount++;
			}
		}
	}
	return TRUE;
}

BOOL CParaStmtInfo::Write(FILE *fp)
{
	//	format for version 1
	//
	//		Stmt	(written by CProcInfo)
	//			id Ver ver
	//			scalar info on one line
	//			m_aSend[0] m_aSendAmount[0] m_aRecv[0] m_aRecvAmount[0]
	//			m_aSend[1] m_aSendAmount[1] m_aRecv[1] m_aRecvAmount[1]
	//			m_aSend[2] m_aSendAmount[2] m_aRecv[2] m_aRecvAmount[2]
	//			...	as many as proc number
	//			EndStmt

	sprintf(szBuf, "%03d %06d \t\t%d %s %d\n", 
		nProcID, nLineCount++, m_nStmtID, KW_VERSION, ms_nVersion);
	fprintf(fp, "%s", szBuf);
	sprintf(szBuf, "%03d %06d \t\t%d %d %d %d %d %d %d %d %d %d %d\n", 
		nProcID, nLineCount++,
		m_nLoopCounter,     
		CPU2usec(m_nElapseCPU),       
		m_nElapseSystem,    
		CPU2usec(m_nTotalElapseCPU),  
		m_nTotalElapseSystem,   
		CPU2usec(m_nComElapseCPU),    
		m_nComElapseSystem, 
		m_nShift,           
		m_nShiftAmount,     
		m_nRemap,           
		m_nRemapAmount);     
	fprintf(fp, "%s", szBuf);

	for (int i=0; i<GetMaxProc(); i++)
	{
		sprintf(szBuf, "%03d %06d \t\t%d %d %d %d\n", nProcID, nLineCount++, 
			m_aSend[i], m_aSendAmount[i], m_aRecv[i], m_aRecvAmount[i]);
		fprintf(fp, "%s", szBuf);
	}
	sprintf(szBuf, "%03d %06d \t\t%s\n", nProcID, nLineCount++, KW_ENDSTMT);
	fprintf(fp, "%s", szBuf);

	return TRUE;
}

BOOL CExecEvent::Read(FILE *fp)
{
	int nVersion;
	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d", 
		&nProcID, &nLineCount, &m_nID, szSymbol, &nVersion);
	
	if (nVersion==1)
	{
		if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
		sscanf(szBuf, "%d %d %d %d %d %d %d %d %d %d %d %d", 
			&nProcID, &nLineCount,
			&m_nStmtID,      
			&m_nState,       
			&m_nTimeCPU,     
			&m_nTimeSystem,  
			&m_nElapseCPU,       
			&m_nElapseSystem,    
			&m_nComTarget,   
			&m_nComSize,     
			&m_nRepCount,    
			&m_nEnclosingID); 

		if (m_nState==stat_para)
			m_nLevel=CExecEvent::LevelIn();
		else if (m_nState==stat_endpara)
			m_nLevel=CExecEvent::LevelOut();
		else m_nLevel=ms_nLevel;
	}
	else if (nVersion==2)
	{
		if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
		sscanf(szBuf, "%d %d %d %d %d %d %d %d %d %d %d %d %d", 
			&nProcID, &nLineCount,
			&m_nStmtID,      
			&m_nState,       
			&m_nTimeCPU,     
			&m_nTimeSystem,  
			&m_nElapseCPU,       
			&m_nElapseSystem,    
			&m_nComTarget,   
			&m_nComSize,     
			&m_nRepCount,    
			&m_nEnclosingID,
			&m_nLevel); 
	}
	return TRUE;
}

BOOL CExecEvent::Write(FILE *fp)
{
	//	format of version1
	//	
	//		Event
	//			id Ver ver
	//			all field on one line, in header file order
	//
	
	sprintf(szBuf, "%03d %06d \t\t%d %s %d\n", nProcID, nLineCount++,
		m_nID, KW_VERSION, ms_nVersion);
	fprintf(fp, "%s", szBuf);
	sprintf(szBuf, "%03d %06d \t\t%d %d %d %d %d %d %d %d %d %d %d\n", 
		nProcID, nLineCount++,
		m_nStmtID,      
		m_nState,       
		CPU2usec(m_nTimeCPU),     
		m_nTimeSystem,  
		CPU2usec(m_nElapseCPU),       
		m_nElapseSystem,    
		m_nComTarget,   
		m_nComSize,     
		m_nRepCount,    
		m_nEnclosingID,
		m_nLevel); 
	fprintf(fp, "%s", szBuf);

	return TRUE;
}

BOOL CProcessors::Read(FILE *fp)
{
	int nVersion;
	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d %d %d %d %d %d %d %d", &nProcID, &nLineCount, 
		&m_nProcessorsID, szSymbol, &nVersion, &m_nGrp, &m_nDimCount,
		&m_aSize[0], &m_aSize[1], &m_aSize[2], &m_aSize[3], &m_aSize[4]); 
	
	if (nVersion==1)
	{
	}
	return TRUE;
}

BOOL CProcessors::Write(FILE *fp)
{
	//	format of version 1
	//
	//		Array	(by CProcInfo)
	//			id Ver ver
	//
	sprintf(szBuf, "%03d %06d \t\t%d %s %d %d %d %d %d %d %d %d\n", 
		nProcID, nLineCount++,
		m_nProcessorsID, KW_VERSION, ms_nVersion, m_nGrp, m_nDimCount,
		m_aSize[0], m_aSize[1], m_aSize[2], m_aSize[3], m_aSize[4]); 
	fprintf(fp, "%s", szBuf);
	return TRUE;
}

BOOL CTemplate::Read(FILE *fp)
{
	int nVersion;
	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d %d %d %d %d %d %d %d", &nProcID, &nLineCount, 
		&m_nTemplateID, szSymbol, &nVersion,
		&m_nDim, &m_nRangeID, &m_nLB, &m_nUB, 
		&m_nDistGroup, &m_nDistType, &m_nBlockSize);
	
	if (nVersion==1)
	{
	}
	return TRUE;
}

BOOL CTemplate::Write(FILE *fp)
{
	//	format of version 1
	//
	//		Array	(by CProcInfo)
	//			id Ver ver
	//
	sprintf(szBuf, "%03d %06d \t\t%d %s %d %d %d %d %d %d %d %d\n", 
		nProcID, nLineCount++,
		m_nTemplateID, KW_VERSION, ms_nVersion, m_nDim, m_nRangeID,
		m_nLB, m_nUB, m_nDistGroup, m_nDistType, m_nBlockSize);
	fprintf(fp, "%s", szBuf);
	return TRUE;
}

BOOL CDistArrayInfo::Read(FILE *fp)
{
	int nVersion;
	if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
	sscanf(szBuf, "%d %d %d %s %d %d %d", &nProcID, &nLineCount, 
		&m_nArrayID, szSymbol, &nVersion, &m_nDAD, &m_nDimAmount);
	
	if (nVersion==1)
	{
		for (int i=0; i<MAX_LEVEL && i<m_nDimAmount; i++)
		{
			if (NULL==fgets(szBuf, sizeof(szBuf)-1, fp)) return FALSE;
			sscanf(szBuf, "%03d %06d \t\t%d %d %d %d %d\n", 
				&nProcID, &nLineCount,
				&m_aLB[i], &m_aUB[i], &m_aAlignRange[i], 
				&m_aAlignType[i], &m_a[i], &m_b[i]);
		}
	}
	return TRUE;
}

BOOL CDistArrayInfo::Write(FILE *fp)
{
	//	format of version 1
	//
	//		Array	(by CProcInfo)
	//			id Ver ver
	//
	sprintf(szBuf, "%03d %06d \t\t%d %s %d %d %d\n", nProcID, nLineCount++,
		m_nArrayID, KW_VERSION, ms_nVersion, m_nDAD, m_nDimAmount);
	fprintf(fp, "%s", szBuf);
	for (int i=0; i<MAX_LEVEL && i<m_nDimAmount; i++)
	{
		sprintf(szBuf, "%03d %06d \t\t%d %d %d %d %d\n", nProcID, nLineCount++,
			m_aLB[i], m_aUB[i], m_aAlignRange[i], 
			m_aAlignType[i], m_a[i], m_b[i]);
		fprintf(fp, "%s", szBuf);
	}
	return TRUE;
}
