//
//	ProcInfo.h
//		define classes of performance data structures
//		include the following:
//			CProcInfo:	container class of all performance data of a processor
//			CParaStmtInfo:	class of all statistical infomation of a parallel
//							statement
//			CExecEvent:	class of event in execution
//			CDistArrayInfo:	class of distributed array 
//

#ifndef _PROCINFO_H
#define _PROCINFO_H

class CParaStmtInfo
{
	//	CParaStmtInfo:	class of statistical infomation of a parallel statement
public:
	static int ms_nVersion;

public:
	CParaStmtInfo();
	CParaStmtInfo(int nStmtID);
	CParaStmtInfo(const CParaStmtInfo &stmt) { *this=stmt; };

public:
	CParaStmtInfo &operator =(const CParaStmtInfo &stmt);

	//	persistent members, recording performance data
public:
	int m_nStmtID;			//	statement ID
	int m_nLoopCounter;		//	times of loops executed
	int m_nElapseCPU;		//	CPU time elapse	
	int m_nElapseSystem;	//	system time elapse	
	int m_nTotalElapseCPU;	//	CPU time elapse, include inner loop
	int m_nTotalElapseSystem;	//	system time elapse, include inner loop
	int m_nComElapseCPU;	//	CPU time elapse for comunication
	int m_nComElapseSystem;	//	system time elapse for communication
	int m_nShift;			//	shift communication count
	int m_nShiftAmount;		//	data amount transferred by shift
	int m_nRemap;			//	remap communication count
	int m_nRemapAmount;		//	data amount transferred by remap
	int m_aSend[MAX_PROC];	//	send op to each processor, one element each
	int m_aSendAmount[MAX_PROC];	//	data amount sent to each processor
	int m_aRecv[MAX_PROC];	//	recv op from each processor, one element each
	int m_aRecvAmount[MAX_PROC];	//	data amount recv from each processor

	//	total value, computed after read
	int m_nSend;
	int m_nSendAmount;
	int m_nRecv;
	int m_nRecvAmount;

	//	temporary data
protected:

	//	methods
public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);
};

class CExecEvent
{
	//	CExecEvent:	class of event in execution
public:
	static int ms_nVersion;

public:
	CExecEvent();
	CExecEvent(int nStmtID);
	CExecEvent(const CExecEvent &event){ *this=event; };

protected:
	static int ms_nNextID;
	int NewID() { ms_nNextID++; return ms_nNextID-1; };
	
	static int ms_nLevel;

public:
	static int LevelIn() { ms_nLevel++; return ms_nLevel; };
	static int LevelOut() { ms_nLevel--; return ms_nLevel; };

public:
	int m_nID;			//	event id, auto generated
	int m_nStmtID;		//	statement id
	int m_nState;		//	execution state tag
	int m_nTimeCPU;		//	CPU time stamp
	int m_nTimeSystem;	//	System time stamp
	int m_nElapseCPU;		//	CPU time stamp, valid only for atomic event
	int m_nElapseSystem;	//	System time stamp, valid only for atomic event

	int m_nComTarget;	//	processor id of communication target
	int m_nComSize;		//	data size of communication
	int m_nRepCount;	//	use this field to process multiple entry of a same 
						//	statement, and reduce the amount of CExecEvent

	int m_nEnclosingID;	//	ID of enclosing loop start event
	int m_nLevel;		//	level of nesting para computing

public:
	CExecEvent &operator =(const CExecEvent &event);

	BOOL EndEvent();

public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);
};

class CProcessors
{
public:
	static int ms_nVersion;

public:
	CProcessors();
	CProcessors(int nTempID);
	CProcessors(const CProcessors &array) { *this=array; };

public:
	CProcessors &operator =(const CProcessors &array);

public:
	int m_nProcessorsID;
	int m_nGrp;
	int m_nDimCount;
	int m_aSize[MAX_LEVEL];

public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);
};

class CTemplate
{
public:
	static int ms_nVersion;

public:
	CTemplate();
	CTemplate(int nTempID);
	CTemplate(const CTemplate &array) { *this=array; };

public:
	CTemplate &operator =(const CTemplate &array);

public:
	int m_nTemplateID;		//	id of template in dep
	int m_nDim;				//	dimension 
	int m_nRangeID;			//	range id
	
	//	template bound
	int m_nLB;		
	int m_nUB;	

	int m_nDistGroup;	//	target of array distribution
	int m_nDistType;	//	BLOCK, CYCLIC, COLLAPSE, COPY
	int m_nBlockSize;	//	block size for (BLOCK CYCLIC) distribution

public:
	BOOL SetDist(int lb, int ub, int nGroup, int nType, int nBlockSize=1);

public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);
};

class CDistArrayInfo
{
	//	CDistArrayInfo:	class of distributed array 
public:
	static int ms_nVersion;

public:
	CDistArrayInfo();
	CDistArrayInfo(const CDistArrayInfo &array) { *this=array; };

public:
	CDistArrayInfo &operator =(const CDistArrayInfo &array);

public:
	int m_nArrayID;		//	id of variable in dep
	int m_nDAD;			//	dad id
	int m_nDimAmount;	//	dimension count
	
	//	array bound
	int m_aLB[MAX_LEVEL];		
	int m_aUB[MAX_LEVEL];	

	//	for array alignment
	int m_aAlignRange[MAX_LEVEL];	//	array align target
	int m_aAlignType[MAX_LEVEL];	//	array align type, NORMAL, COLLAPSE, COPY
	int m_a[MAX_LEVEL], m_b[MAX_LEVEL];		//	Align parameter, aX+b

public:
	SetArrayAlign(int nDim, int nLB, int nUB,
		int a, int b, int nRange, int nType=1);
	
public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);
};

class CProcInfo
{
	//	CProcInfo:	container class of all performance data of a processor
public:
	static int ms_nVersion;

public:
	CProcInfo();

	//	persistent members, recording performance data
public:
	int m_nProcID;			//	processor id 
	int m_nTimeCPU;			//	start up CPU time stamp
	int m_nTimeSystem;		//	start up system time stamp

	int m_nElapseCPU;		//	elapse of CPU time
	int m_nElapseSystem;	//	elapse of system time

	int m_aSend[MAX_PROC];	//	send op to each processor, one element each
	int m_aSendAmount[MAX_PROC];	//	data amount sent to each processor
	int m_aRecv[MAX_PROC];	//	recv op from each processor, one element each
	int m_aRecvAmount[MAX_PROC];	//	data amount recv from each processor

public:
	vector <CParaStmtInfo> m_aStmts;		//	array of parallel statments
	vector <CExecEvent> m_aEvents;			//	array of execution events

	vector <CDistArrayInfo> m_aDistArrays;	//	array of distributed arrays
	vector <CTemplate> m_aTemplates;	//	array of distributed arrays
	vector <CProcessors> m_aProcDefs;	//	array of distributed arrays

	//	manipulation methods
public:
	BOOL Start(int nProcID);	//	start collecting profile data
	BOOL Stop();				//	stop collecting profile data

	BOOL GetSendTotal(int &nCount, int &nAmount);
	BOOL GetRecvTotal(int &nCount, int &nAmount);

	CParaStmtInfo *NewStmt(int nStmtID);
	CParaStmtInfo *FindStmt(int nStmtID);

	POSITION GetFirstStmtPos();
	POSITION GetLastStmtPos();
	CParaStmtInfo *GetStmt(POSITION pos);
	CParaStmtInfo *GetNextStmt(POSITION &pos);
	CParaStmtInfo *GetPrevStmt(POSITION &pos);

	CExecEvent *NewEvent(int nStmtID, int nState, 
		int nComTarget=0, int nComSize=0);
	BOOL EraseLastEvent();

	POSITION GetFirstEventPos(int nStmtID=-1);
	POSITION GetLastEventPos(int nStmtID=-1);
	CExecEvent *GetEvent(POSITION pos);
	CExecEvent *GetNextEvent(POSITION &pos, int nStmtID=-1);
	CExecEvent *GetPrevEvent(POSITION &pos, int nStmtID=-1);
	
	POSITION FindEventByTime(int nTimeCPU);
	POSITION FindEventBySysTime(int nTimeSystem);
	POSITION FindEventByTime(int nTimeCPU, int nSpan, int &posNext);
	POSITION FindEventBySysTime(int nTimeSystem, int nSpan, int &posNext);
	BOOL GetStatus(int &nStatus, POSITION pos, POSITION posNext);
	BOOL GetStatusByTime(int nTimeCPU, int nSpan, 
		int &nStatus, POSITION &pos);
	BOOL GetStatusBySysTime(int nTimeSystem, int nSpan,
		int &nStatus, POSITION &pos);
	BOOL GetCommByTime(int nTimeCPU, int nSpan, 
		int aSend[MAX_PROC], int aSendAmount[MAX_PROC]);
	BOOL GetCommBySysTime(int nTimeSystem, int nSpan,
		int aSend[MAX_PROC], int aSendAmount[MAX_PROC]);
	
	CProcessors *NewProcDef(int nProcessorsID, int nGroupID);
	CProcessors *FindProcDef(int nProcessorsID);
	CProcessors *FindProcDefByGrp(int nGroupID);

	CTemplate *NewTemplate(int nTemplateID, int nRangeID, int nDim);
	CTemplate *FindTemplate(int nTemplateID, int nDim);
	CTemplate *FindTemplateByRange(int nRangeID);

	CDistArrayInfo *NewDistArray(int nArrayID, int nDAD);
	CDistArrayInfo *FindDistArray(int nArrayID);
	CDistArrayInfo *FindDistArrayByDAD(int nDAD);

	POSITION GetFirstArrayPos();
	POSITION GetLastArrayPos();
	CDistArrayInfo *GetArray(POSITION pos);
	CDistArrayInfo *GetNextArray(POSITION &pos);
	CDistArrayInfo *GetPrevArray(POSITION &pos);

public:		//	input & output
	BOOL Read(FILE *fp);
	BOOL Write(FILE *fp);

	//	temporary data
protected:

};

#endif
